"""An objective function for minimizing the makespan of `Gantt` charts."""
from typing import Final

import numba  # type: ignore

from moptipy.api.objective import Objective
from moptipy.examples.jssp.gantt import Gantt
from moptipy.examples.jssp.instance import Instance
from moptipy.utils.types import type_error


# start book
@numba.njit(nogil=True, cache=True)
def makespan(x: Gantt) -> int:
    """
    Get the makespan corresponding to a given `Gantt` chart.

    The makespan corresponds to the maximum of the end times of the
    last operation on each machine. This is jitted for performance.

    :param x: the Gantt chart.
    :return: the maximum of any end time stored in the chart
    """
    return int(x[:, -1, 2].max())  # maximum of end time of last op
    # end book


# start book
class Makespan(Objective):
    """Compute the makespan of a `Gantt` chart (for minimization)."""

# end book
    def __init__(self, instance: Instance) -> None:  # +book
        """
        Initialize the makespan objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        self.__instance: Final[Instance] = instance
        #: The fast call forwarding to the makespan function. # +book
        self.evaluate = makespan  # type: ignore # +book

    def lower_bound(self) -> int:
        """
        Get the lower bound of the makespan.

        :return: the lower bound
        """
        return self.__instance.makespan_lower_bound

    def is_always_integer(self) -> bool:
        """
        Return `True` because :func:`makespan` always returns `int` values.

        :retval True: always
        """
        return True

    def upper_bound(self) -> int:
        """
        Get the upper bound of the makespan.

        :return: the sum of all job execution times.
        """
        return self.__instance.makespan_upper_bound

    def __str__(self) -> str:
        """
        Get the name of the makespan objective function.

        :return: `makespan`
        :retval "makespan": always
        """
        return "makespan"



"""
An operator trying to change exactly `n` elements in a permutation.

This is an operator with a step size
(:class:`moptipy.api.operators.Op1WithStepSize`) and the step size determines
how many elements of a permutation should be changed.

Let's say we have a "normal" permutation where each element occurs once.
The permutation has length `n`, let's say that its
:attr:`~moptipy.spaces.permutations.Permutations.blueprint` is
`[0, 1, 2, 3, 4, ..., n-1]`. Then, with one application of the operator, we
can change no less than two elements (`step_size=0.0`) and no more than `n`
(`step_size=1.0`). Clearly we cannot change only one element, because what
different value could we put there? We would violate the "permutation nature."
We can also not change more than `n` elements, obviously.

What our operator does is then it computes `m` by using
:func:`~moptipy.operators.tools.exponential_step_size`.
This ensures that `m=2` for `step_size=0.0` and `m` is maximal for
`step_size=1.0`. In between the two, it extrapolates exponentially. This means
that small values of `step_size` will lead to few swap moves, regardless of
the length of the permutation and many swaps are performed only for
`step_size` values close to `1`.
Then it will pick `m` random indices and permute the elements at them such
that none remains at its current position.
That is rather easy to do.
Unfortunately, :class:`~moptipy.spaces.permutations.Permutations` allows also
permutations with repetitions. Here, things get dodgy.

Let's say we have the blueprint `[0, 1, 1, 1, 1]`. Then we can only change
exactly two elements.

>>> print(get_max_changes([0, 1, 1, 1, 1]))
2

If we have the blueprint `[0, 0, 1, 1, 1]`, then we can change two elements.
We can also change four elements. But there is no way to change three elements!

>>> print(get_max_changes([0, 0, 1, 1, 1]))
4

So in the general case, we determine the maximum number `mx` of positions that
can be changed with one move via the function :func:`get_max_changes`.
We can still translate a `step_size` to a number `m` by doing
`m = 2 + int(round(step_size * (mx - 2)))`.
However, it is not clear whether all moves from `m=2` to `m=mx` are actually
possible.

And even if they are possible, they might be very hard to sample randomly.
Some moves touching a large number `m` of positions may be restricted to
swap the elements at very specific indicies in a very specific order. Finding
such a move by chance may be rather unlikely.

Indeed, a search operator should do random moves. I did not find a way to
construct moves according to a policy which is both random and can yield any
move.
So I divide the operator into two steps:

First, the function :func:`find_move` tries to find a move for a given `m`
in a random fashion. This function tries to find the sequence of indices for a
cyclic swap where each element is different from both its successor and
predecessor. It builds such index sequences iteratively. This may fail:
as in the example above, for some values of `m` it might just not be
possible to construct a suitable sequence, either because there is none or
because building it randomly has too low of a chance of success. Hence, the
function tries to build at most `max_trials` sequences. Whenever building a
sequence, it also remembers the longest-so-far cyclic move. If that one is
shorter than `m` but all `max_trials` trials are exhausted, it returns this
sequence instead. So in summary, this function tries to find the longest
cyclic swap which is not longer than `m` and returns it. It may be shorter
than `m`. If we deal with permutations where each value occurs only once, this
function is guaranteed to find a sequence of length `m` in the first trial,
i.e., it does not waste runtime.

Once a move was found, the function :func:`apply_move` applies it. Now, as
said, :func:`find_move` discovers cyclic changes that are reasonably random.
However, cyclic changes are not the only possible moves of length `m`. For
example, if we have the permutation blueprint `[0, 1, 2, 3, 4]`, a move of
length 4 could be to exchange `0` with `1` and to swap `2` with `3`.
:func:`find_move` cannot find this move, but it could find a cyclic swap of
`0` to `1`, `1` to `2`, `2` to `3`, and `3` to `1`. So it could find the
right indices for such a move, just restricted to the cyclic swapping. So
what :func:`apply_move` tries to do is to permute the indices discovered by
:func:`find_move` randomly and check whether this would still yield a feasible
move changing exactly `m` locations. Any shuffling of the elements at the
selected position which avoids putting the original values into the original
positions would do. Sometimes, most such shuffles work out, e.g., if we work
on the space of permutations where each element occurs once. Sometimes, the
only sequence that works is indeed the cyclic move and shuffling it cannot
work. So :func:`apply_move` again has a limit for the maximum number of
attempts to find a shuffle that works out. If it finds one, then it applies
it as move. If it does not find one and the trials are exhausted, it randomly
choses whether to apply the cyclic move as a cycle to the left or as a cycle
to the right. Either way, it will change exactly `m` positions of the
permutation, as prescribed by the move. Cycling one step in either direction
will always work, since each element is different from both its predecessor
and successor. (Cycling more than one step (but less than `m`) could
potentially fail in permutations with repetitions, because there is no
guarantee that any element is different from its successor's successor.)

Now the overall operator just plugs these two functions together. It also adds
one slight improvement: If we demand to change a number `q` of locations for
the first time and :func:`find_move` fails to find a move of length `q` but
instead offers one of length `p<q`, then the operator remembers this. The next
time we ask to change `q` positions, it will directly try to change only `p`.
This memory is reset in :meth:`~Op1SwapExactlyN.initialize`.
"""
from typing import Callable, Counter, Final, Iterable

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1WithStepSize
from moptipy.operators.tools import exponential_step_size
from moptipy.spaces.permutations import Permutations
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_INT, fill_in_canonical_permutation
from moptipy.utils.types import check_int_range, type_error


def get_max_changes(blueprint: Iterable[int]) -> int:
    """
    Get the maximum number of changes possible for a given permutation.

    :param blueprint: the blueprint permutation
    :returns: the maximum number of changes possible

    >>> get_max_changes("1")
    0
    >>> get_max_changes("11")
    0
    >>> get_max_changes("12")
    2
    >>> get_max_changes("123")
    3
    >>> get_max_changes("1233")
    4
    >>> get_max_changes("12333")
    4
    >>> get_max_changes("1234")
    4
    >>> get_max_changes("12344")
    5
    >>> get_max_changes("123444")
    6
    >>> get_max_changes("1234444")
    6
    >>> get_max_changes("12334444")
    8
    >>> get_max_changes("123344445")
    9
    >>> get_max_changes("1233444455")
    10
    >>> get_max_changes("12334444555")
    11
    >>> get_max_changes("123344445555")
    12
    >>> get_max_changes("1233444455555")
    13
    >>> get_max_changes("12334444555555")
    14
    >>> get_max_changes("123344445555555")
    15
    >>> get_max_changes("1233444455555555")
    16
    >>> get_max_changes("12334444555555555")
    16
    >>> get_max_changes("112233")
    6
    >>> get_max_changes("11223344")
    8
    >>> get_max_changes("112233445")
    9
    >>> get_max_changes("1122334455")
    10
    >>> get_max_changes("11223344555")
    11
    >>> get_max_changes("112233445555555")
    15
    >>> get_max_changes("1122334455555555")
    16
    >>> get_max_changes("11223344555555555")
    16
    """
    # Create tuples of (count, negative priority, value).
    counts: Final[list[list[int]]] = [
        [a[1], 0, a[0]] for a in Counter[int](blueprint).most_common()]
    if len(counts) <= 1:
        return 0

# We simulate a chain swap. We begin with the element that appears the least
# often. We take this element and reduce its count. Next, we take the element
# that appears the most often. We reduce its count. And so on. Ties are broken
# by prioritizing the element that was not used the longest. If the element
# that we want to pick is the same as the previously picked one, we skip over
# it. If no element can be picked, we quit.
    changes: int = 0
    last: int | None = None
    found: bool = True
    smallest: bool = True
    while found:
        found = False
# We sort alternatingly sometimes we pick the smallest, sometimes the largest.
        if smallest:
            counts.sort()
        else:
            counts.sort(key=lambda a: [-a[0], a[1]])
        smallest = not smallest  # realize the alternating sorting
# Now we iterate over the sorted list and pick the first suitable element.
# I know: There will be at most two iterations of this loop ... but whatever.
        for idx, chosen in enumerate(counts):
            current = chosen[2]
            if current == last:  # We cannot pick the same element twice
                continue         # in a row. So we skip over it.
            cnt = chosen[0]
            if cnt <= 1:  # How often does this element exist in the list?
                del counts[idx]  # No more elements of this type.
            else:
                chosen[0] = cnt - 1  # There is at least one more left over.
            changes = changes + 1  # Yeah, we can increase the changes.
            chosen[1] = changes  # Increase negative priority.
            found = True  # We found an element.
            last = current  # Remember the element.
            break  # Quit (after at most two loop iterations).
    if changes <= 1:  # This cannot be.
        raise ValueError(
            f"Error in counting possible changes for {blueprint}.")
    return changes


def find_move(x: np.ndarray, indices: np.ndarray, step_size: int,
              ri: Callable[[int, int], int], max_trials: int,
              temp: np.ndarray) -> np.ndarray:
    """
    Find a move of at most step_size changes and store it in indices.

    For any pure permutation `x`, we can basically find any move between
    `step_size=2` and `len(x)`. If we have permutations with repetitions, the
    upper limit becomes smaller, because it may not be possible to change all
    elements at once. For example, in `[0, 0, 1, 1, 1, 1, 1]`, we can change
    at most four elements. For this purpose, :func:`get_max_changes` exists.
    It tells us the upper limit of changes, which would be four here:

    >>> print(get_max_changes([0, 0, 1, 1, 1, 1, 1]))
    4

    Yet, this does not necessarily mean that we can make all moves including
    2, 3, and 4 changes. If you look at the above example, you may find that
    it is impossible to change exactly 3 locations. We can only actually do
    step sizes of 2 and 4, but neither 3 nor 5, 6, nor 7.
    In reality, even if it was possible to make a big move, it might be very
    unlikely to be able to this in a *random* fashion. If you look at the
    code of :func:`get_max_changes`, it tries to find the longest possible
    cyclic move by working through the elements of the permutation in a very
    specific order. Using a different (or random) order would yield shorter
    moves.

    Our method :func:`find_move` here therefore tries to find the longest
    possible cyclic swap involving at most `step_size` indices. It may find
    such a move that uses exactly `step_size` indices. But it may also find
    a shorter move because either the perfect complete feasible move touching
    `step_size` indices does not exist or because finding it in a random
    fashion may take too long. (For this purpose, `max_trials` exists to
    limit the search for the right move.)

    For normal permutations, it will find the move of exactly the right
    length. For permutations with repetitions, it has a decent chance to find
    a move of the length `step_size` if a) such a move exists and
    b) `step_size` is not too big. Otherwise, it should find a shorter but
    still reasonably large move.

    :param x: the permutation to be modified in the end
    :param indices: the set of indices, which must be the canonical
        permutation of `0...len(x)-1`
    :param step_size: the step size, i.e., the number of elements to change
    :param ri: a function to get random integers
    :param max_trials: the maximum number of trials
    :param temp: a temporary array
    :returns: the discovered move

    >>> import numpy as npx
    >>> from numpy.random import default_rng
    >>> rand_int = default_rng(12).integers
    >>> perm = npx.array([0, 1, 2, 3, 3, 3, 3, 3, 3], int)
    >>> use_indices = npx.array(range(len(perm)), int)
    >>> want_size = len(perm)
    >>> print(want_size)
    9
    >>> print(get_max_changes(perm))
    6
    >>> use_temp = npx.empty(len(perm), int)
    >>> res = find_move(perm, use_indices, want_size, rand_int, 1000, use_temp)
    >>> print(res)
    [5 2 8 1 4 0]
    >>> perm2 = perm.copy()
    >>> perm2[res] = perm2[npx.roll(res, 1)]
    >>> print(f"{perm} vs. {perm2}")
    [0 1 2 3 3 3 3 3 3] vs. [3 3 3 3 1 0 3 3 2]
    >>> print(sum(perm != perm2))
    6
    >>> perm = npx.array([1, 3, 5, 9, 4, 2, 11, 7], int)
    >>> want_size = len(perm)
    >>> print(want_size)
    8
    >>> use_indices = npx.array(range(len(perm)), int)
    >>> use_temp = npx.empty(len(perm), int)
    >>> res = find_move(perm, use_indices, want_size, rand_int, 1, use_temp)
    >>> print(res)
    [4 0 5 3 1 7 2 6]
    >>> print(len(res))
    8
    >>> perm2 = perm.copy()
    >>> perm2[res] = perm2[npx.roll(res, 1)]
    >>> print(f"{perm} vs. {perm2}")
    [ 1  3  5  9  4  2 11  7] vs. [ 4  9  7  2 11  1  5  3]
    >>> print(sum(perm != perm2))
    8
    """
    length: Final[int] = len(x)
    lm1: Final[int] = length - 1
    sm1: int = step_size - 1
    best_size: int = -1

# We spent at most `max_trials` trials to find a perfect move. The reason is
# that some values of `step_size` might be impossible to achieve, other may
# be achievable only with a very low probability. So we need a trial limit to
# avoid entering a very long or even endless loop.
    for _ in range(max_trials):
        current_best_size: int = -1  # longest feasible move this round
# We aim to store a suitable selection of indices at indices[0:step_size].
# Like in :mod:`moptipy.operators.permutations.op1_swapn`, we try to create a
# feasible move as a series of cyclic swaps. Basically, the element at index
# i1 would be moved to index i2, the element at i2 to i3, ..., the element at
# the last index ix to i1. We make sure that all indices are used at most
# once.
# We proceed similar to a partial Fisher-Yates shuffle.
# The variable "found" is the number of indices that we have fixed so far.
# For each new position, we test the remaining indices in random order.
# An index can only be used if it points to an element different to what the
# last index points to, or else the cyclic swap would be meaningless.
# For the last index to be chosen, that element must also be different from
# the first element picked.
# So a potential index may not work. We test each index at most once by moving
# the tested indices into the range indices[found:tested] and keep increasing
# tested and reset it once we found a new acceptable index.
# This procedure, however, may end up in a dead end. We therefore need to wrap
# it into a loop that tries until success...  ...but we cannot do that because
# success may not be possible. Some moves may be impossible to do.
        i_idx: int = ri(0, length)  # Get the first random index.
        i_src: int = indices[i_idx]  # Pick the actual index at that index.
        indices[0], indices[i_idx] = i_src, indices[0]  # Swap it to 0.
        found: int = 1  # the number of found indices
        tested: int = 1  # the number of tested indices
        last = first = x[i_src]  # Get the value at the first index.
        continue_after: bool = found < sm1  # continue until step_size-1
        while True:
            i_idx = ri(tested, length)  # Get random index.
            i_dst: int = indices[i_idx]  # Pick the actual index.
            current: int = x[i_dst]  # Get value at the new index.
            can_end: bool = current != first
            accept: bool = (current != last) and (continue_after or can_end)
            if accept:
                # Now we move the new index into the "found" section.
                indices[found], indices[i_idx] = i_dst, indices[found]
                tested = found = found + 1  # found++, reset tested.
                last = current  # Store value from i_dst in last.
                if can_end:
                    current_best_size = found
                    if not continue_after:
                        return indices[0:found]  # We won!
                continue_after = found < sm1  # continue until step_size-1
                continue
            if tested >= lm1:  # Are we in a dead end?
                break  # and try again freshly.
            indices[tested], indices[i_idx] = i_dst, indices[tested]
            tested = tested + 1  # We have tested another value.

# If we get here, we did not succeed in creating a complete move this round.
# However, current_best_size should indicate one possible shorter move.
# It must be at least >= 2, because we will definitely be able to find two
# different elements in x. We will remember the longest feasible move
# discovered in all max_trials iterations.
        if current_best_size > best_size:
            best_size = current_best_size
            temp[0:best_size] = indices[0:best_size]

# If get here, we have completed the main loop. We could not find a complete
# feasible move that changes exactly step_size locations. However, we should
# have a shorter move by now stored in temp. So we copy its indices.
    return temp[0:best_size]


def apply_move(x: np.ndarray, dest: np.ndarray, move: np.ndarray,
               shuffle: Callable[[np.ndarray], None],
               ri: Callable[[int, int], int], max_trials: int) -> None:
    """
    Apply a given move copying from source `x` to `dest`.

    `move` must be the return value of :func:`find_move`.

    :param x: the source permutation
    :param dest: the destination permutation
    :param move: the move of length `step_size`
    :param shuffle: a source for random shuffling
    :param ri: a source for random integers
    :param max_trials: a maximum number of trials

    >>> import numpy as npx
    >>> from numpy.random import default_rng
    >>> random = default_rng(12)
    >>> shuff = random.shuffle
    >>> rint = random.integers
    >>> xx = npx.array([0, 1, 2, 3, 3, 3, 3, 3, 3], int)
    >>> dd = npx.empty(len(xx), int)
    >>> mv = npx.array([5, 2, 8, 1, 4, 0], int)
    >>> print(len(mv))
    6
    >>> apply_move(xx, dd, mv, shuff, rint, 0)
    >>> print(dd)
    [3 3 3 3 0 2 3 3 1]
    >>> print(sum(dd != xx))
    6
    >>> apply_move(xx, dd, mv, shuff, rint, 1000)
    >>> print(dd)
    [3 3 3 3 2 1 3 3 0]
    >>> print(sum(dd != xx))
    6
    >>> xx = npx.array([1, 3, 5, 9, 4, 2, 11, 7], int)
    >>> dd = npx.empty(len(xx), int)
    >>> mv = npx.array([4, 0, 5, 3, 1, 7, 2, 6], int)
    >>> print(len(mv))
    8
    >>> apply_move(xx, dd, mv, shuff, rint, 0)
    >>> print(dd)
    [ 4  9  7  2 11  1  5  3]
    >>> print(sum(dd != xx))
    8
    >>> xx = npx.array([1, 3, 5, 9, 4, 2, 11, 7], int)
    >>> apply_move(xx, dd, mv, shuff, rint, 10)
    >>> print(dd)
    [ 5  4  2 11  7  3  9  1]
    >>> print(sum(dd != xx))
    8
    """
    np.copyto(dest, x)  # First, we copy `x` to `dest`.
    step_size: Final[int] = len(move)  # Get the move size.
# We now try to replace the sub-string x[move] with a random permutation of
# itself. Since `move` is the return value of `find_move`, we know that it
# could be realized as a cyclic swap. However, we are not yet sure whether it
# can also be some other form of move, maybe swapping a few pairs of elements.
# We try to find this out in this loop by testing `max_trials` random
# permutations of the original subsequence.
    orig: np.ndarray = dest[move]   # Get the original subsequence.
    perm: np.ndarray = orig.copy()
    for _ in range(max_trials):  # Try to permute it.
        shuffle(perm)  # Shuffle the permutation
        if sum(perm != orig) == step_size:  # If all elements are now...
            dest[move] = perm  # ...at different places, accept the move and
            return  # quit
# If we get here, trying random permutations of the sequence did not work.
# We now simply perform a cyclic swap. A cyclic swap has the property that
# each element is different from both its predecessor and its successor.
# Hence, we could cycle left or cycle right. Now one could argue that this
# should not matter, since we have constructed the move in a random way via
# find_move. However, I am not sure. If we have a permutation with repetition,
# then finding certain move patterns could be more likely than others. For
# example, if one element occurs much more often than another one, one of its
# instance could be more likely chosen as first element of the move. A more
# rarely occurring element could then appear later in the move when there are
# not many other choices left. Hence, it might be that the moves found by
# find_move are not uniformly distributed over the realm of possible moves of
# a given step length. In this case, it may be useful to randomly choose
# whether to cycle left or right. We cannot cycle more than one step, though,
# because there is guarantee that an element is different from its successor's
# successor.
    dest[move] = np.roll(orig, 1 if ri(0, 2) == 0 else -1)


class Op1SwapExactlyN(Op1WithStepSize):
    """A unary search operator that swaps `n` (different) elements."""

    def __init__(self, perm: Permutations,
                 max_move_trials: int = 1000,
                 max_apply_trials: int = 1000) -> None:
        """
        Initialize the operator.

        :param perm: the base permutation
        :param max_move_trials: the maximum number of attempts to generate a
            fitting move
        :param max_apply_trials: the maximum number of attempts to apply a
            move via a random permutation
        """
        super().__init__()
        if not isinstance(perm, Permutations):
            raise type_error(perm, "perm", Permutations)

# If each value appears exactly once in the permutation, then we can
# easily do perm.dimension changes. If the permutation is with
# repetitions, then it might be fewer, so we need to check.
        is_pure_perm: Final[bool] = perm.n() == perm.dimension
        #: the maximum number of possible changes
        self.max_changes: Final[int] = check_int_range(
            perm.dimension if is_pure_perm
            else get_max_changes(perm.blueprint),
            "max_changes", 2, 100_000_000)
        #: the maximum number of attempts to find a move with the exact step
        #: size
        self.max_move_trials: Final[int] =\
            check_int_range(max_move_trials, "max_move_trials", 1)
        #: the maximum number of attempts to apply a random permutation of the
        #: move before giving up and applying it as cyclic swap
        self.max_apply_trials: Final[int] =\
            check_int_range(max_apply_trials, "max_apply_trials", 1)
        #: the set of chosen indices
        self.__indices: Final[np.ndarray] = np.empty(
            perm.dimension, DEFAULT_INT)
        #: a temporary array
        self.__temp: Final[np.ndarray] = np.empty(
            perm.dimension, DEFAULT_INT)
        #: the initial move map that maps step sizes to feasible sizes
        self.__initial_move_map: Final[dict[int, int]] = {i: i for i in range(
            2, perm.dimension + 1)} if is_pure_perm else {2: 2}
        #: the move map that maps step sizes to feasible sizes
        self.__move_map: Final[dict[int, int]] = {}

    def initialize(self) -> None:
        """Initialize this operator."""
        super().initialize()
        fill_in_canonical_permutation(self.__indices)
        self.__move_map.clear()
        self.__move_map.update(self.__initial_move_map)

    def op1(self, random: Generator, dest: np.ndarray, x: np.ndarray,
            step_size: float = 0) -> None:
        """
        Copy `x` into `dest` and then swap several different values.

        :param random: the random number generator
        :param dest: the array to receive the modified copy of `x`
        :param x: the existing point in the search space
        :param step_size: the number of elements to swap
        """
# setup of other variables and fast calls
        ri: Final[Callable[[int, int], int]] = random.integers  # fast call!
# compute the real step size based on the maximum changes
        use_step_size: int = exponential_step_size(
            step_size, 2, self.max_changes)
# look up in move map: Do we already know whether this step size works or can
# it be replaced with a similar smaller one?
        move_map: Final[dict[int, int]] = self.__move_map
        mapped_step_size = move_map.get(use_step_size, -1)
        if mapped_step_size >= 2:
            use_step_size = mapped_step_size

        move: Final[np.ndarray] = find_move(
            x, self.__indices, use_step_size, ri, self.max_move_trials,
            self.__temp)

# If we did not yet know the move size, then update it.
        if mapped_step_size == -1:
            self.__move_map[use_step_size] = len(move)
        self.__move_map[use_step_size] = use_step_size
# Finally, apply the move.
        apply_move(x, dest, move, random.shuffle, ri, self.max_move_trials)

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "swapxn", the name of this operator
        :retval "swapxn": always
        """
        return "swapxn"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this operator to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("maxChanges", self.max_changes)
        logger.key_value("maxMoveTrials", self.max_move_trials)
        logger.key_value("maxApplyTrials", self.max_apply_trials)

# Translate the move map to some strings like "2;5-8;12", meaning that all the
# moves 2, 5, 6, 7, 8, and 12 were performed and valid.
        kvs: list[int] = [k for k, v in self.__move_map.items() if k == v]
        kvs.sort()
        lk: Final[int] = len(kvs) - 1
        kvs = [(e if ((i <= 0) or (i >= lk) or kvs[i - 1] != (e - 1))
                or (kvs[i + 1] != (e + 1)) else -1)
               for i, e in enumerate(kvs)]
        kvs = [e for i, e in enumerate(kvs) if
               (i <= 0) or (i >= lk) or (e != -1) or (kvs[i - 1] != -1)]
        logger.key_value("moves", CSV_SEPARATOR.join(
            "-" if k < 0 else str(k) for k in kvs).replace(
            f"{CSV_SEPARATOR}-{CSV_SEPARATOR}", "-"))



"""An implementation of an integer string based search space."""

from typing import Final

import numpy as np

from moptipy.spaces.nparrayspace import NPArraySpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.strings import num_to_str_for_name
from moptipy.utils.types import type_error

#: the log key for the minimum value
KEY_MIN: Final[str] = "min"
#: the log key for the maximum value
KEY_MAX: Final[str] = "max"


class IntSpace(NPArraySpace):
    """
    A space where each element is a one-dimensional numpy integer array.

    Such spaces can serve as basis for implementing combinatorial
    optimization and can be extended to host permutations. Their elements are
    instances of :class:`numpy.ndarray`.

    >>> s = IntSpace(5, -4, 99)
    >>> print(s.dimension)
    5
    >>> print(s.min_value)
    -4
    >>> print(s.max_value)
    99
    >>> print(s.dtype)
    int8
    >>> s = IntSpace(5, 2, 200)
    >>> print(s.dtype)
    uint8
    >>> s = IntSpace(5, 2, 202340)
    >>> print(s.dtype)
    int32
    """

    def __init__(self, dimension: int,
                 min_value: int, max_value: int) -> None:
        """
        Create the integer-based search space.

        :param dimension: The dimension of the search space,
            i.e., the number of decision variables.
        :param min_value: the minimum value
        :param max_value: the maximum value
        """
        if not isinstance(min_value, int):
            raise type_error(min_value, "min_value", int)
        if not isinstance(max_value, int):
            raise type_error(max_value, "max_value", int)
        if min_value >= max_value:
            raise ValueError(
                f"max_value > min_value must hold, but got "
                f"min_value={min_value} and max_value={max_value}.")
        super().__init__(dimension, int_range_to_dtype(
            min_value=min_value, max_value=max_value))
        #: the lower bound, i.e., the minimum permitted value
        self.min_value: Final[int] = min_value
        #: the upper bound, i.e., the maximum permitted value
        self.max_value: Final[int] = max_value

    def create(self) -> np.ndarray:
        """
        Create an integer vector filled with the minimal value.

        :return: the vector

        >>> from moptipy.spaces.intspace import IntSpace
        >>> s = IntSpace(dimension=12, min_value=5, max_value=332)
        >>> v = s.create()
        >>> print(s.to_str(v))
        5;5;5;5;5;5;5;5;5;5;5;5
        >>> print(v.dtype)
        int16
        """
        return np.full(shape=self.dimension, fill_value=self.min_value,
                       dtype=self.dtype)

    def validate(self, x: np.ndarray) -> None:
        """
        Validate an integer string.

        :param x: the integer string
        :raises TypeError: if the string is not an element of this space.
        :raises ValueError: if the shape of the vector is wrong or any of its
            element is not finite.
        """
        super().validate(x)

        miv: Final[int] = self.min_value
        mav: Final[int] = self.max_value
        for index, item in enumerate(x):
            if not (miv <= item <= mav):
                raise ValueError(
                    f"x[{index}]={item}, but should be in {miv}..{mav}.")

    def n_points(self) -> int:
        """
        Get the number of possible different integer strings.

        :return: (max_value - min_value + 1) ** dimension

        >>> print(IntSpace(4, -1, 3).n_points())
        625
        """
        return (self.max_value - self.min_value + 1) ** self.dimension

    def __str__(self) -> str:
        """
        Get the name of this integer space.

        :return: "ints" + dimension + dtype.char + min_value + "-" + max_value

        >>> print(IntSpace(4, -1, 3))
        ints4bm1to3
        """
        return f"ints{self.dimension}{self.dtype.char}"\
               f"{num_to_str_for_name(self.min_value)}to"\
               f"{num_to_str_for_name(self.max_value)}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this space to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> space = IntSpace(7, -5, 5)
        >>> space.dimension
        7
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         space.log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[-2]
        'max: 5'
        >>> text[-3]
        'min: -5'
        >>> text[-4]
        'dtype: b'
        >>> text[-5]
        'nvars: 7'
        >>> text[-6]
        'class: moptipy.spaces.intspace.IntSpace'
        >>> text[-7]
        'name: ints7bm5to5'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        logger.key_value(KEY_MIN, self.min_value)
        logger.key_value(KEY_MAX, self.max_value)



"""
The base class for implementing objective functions.

An objective function evaluates the quality of a candidate solution of an
optimization problem. Solutions with smaller objective values are better,
i.e., objective functions are subject to minimization. All objective
functions inherit from :class:`~moptipy.api.objective.Objective`. If you
implement a new objective function, you can test it via the pre-defined unit
test routine :func:`~moptipy.tests.objective.validate_objective`.
"""
from math import inf, isfinite
from typing import Any

from moptipy.api import logging
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error


# start book
class Objective(Component):
    """
    An objective function subject to minimization.

    An objective function represents one optimization criterion that
    is used for rating the solution quality. All objective functions in
    our system are subject to minimization, meaning that smaller values
    are better.
    """

    def evaluate(self, x) -> float | int:
        """
        Evaluate a solution `x` and return its objective value.

        The return value is either an integer or a float and must be
        finite. Smaller objective values are better, i.e., all objective
        functions are subject to minimization.

        :param x: the candidate solution
        :return: the objective value
        """
    # end book

    def lower_bound(self) -> float | int:
        """
        Get the lower bound of the objective value.

        This function returns a theoretical limit for how good a solution
        could be at best. If no real limit is known, the function should
        return `-inf`.

        :return: the lower bound of the objective value
        """
        return -inf

    def upper_bound(self) -> float | int:
        """
        Get the upper bound of the objective value.

        This function returns a theoretical limit for how bad a solution could
        be at worst. If no real limit is known, the function should return
        `inf`.

        :return: the upper bound of the objective value
        """
        return inf

    def is_always_integer(self) -> bool:
        """
        Return `True` if :meth:`~evaluate` will always return an `int` value.

        :returns: `True` if :meth:`~evaluate` will always return an `int`
          or `False` if also a `float` may be returned.
        """
        return False

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this function to the provided destination.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        b = self.lower_bound()
        if isinstance(b, int) or isfinite(b):
            logger.key_value(logging.KEY_F_LOWER_BOUND, b)
        b = self.upper_bound()
        if isinstance(b, int) or isfinite(b):
            logger.key_value(logging.KEY_F_UPPER_BOUND, b)


def check_objective(objective: Any) -> Objective:
    """
    Check whether an object is a valid instance of :class:`Objective`.

    :param objective: the objective
    :return: the objective
    :raises TypeError: if `objective` is not an instance of
        :class:`Objective`

    >>> check_objective(Objective())
    Objective
    >>> try:
    ...     check_objective('A')
    ... except TypeError as te:
    ...     print(te)
    objective function should be an instance of moptipy.api.objective.\
Objective but is str, namely 'A'.
    >>> try:
    ...     check_objective(None)
    ... except TypeError as te:
    ...     print(te)
    objective function should be an instance of moptipy.api.objective.\
Objective but is None.
    """
    if isinstance(objective, Objective):
        return objective
    raise type_error(objective, "objective function", Objective)



"""An archive of solutions to a multi-objective problems."""

from typing import Any, Final

import numpy as np

from moptipy.api.component import Component
from moptipy.api.mo_utils import lexicographic
from moptipy.utils.nputils import array_to_str
from moptipy.utils.types import type_error


class MORecord:
    """
    A record for the multi-objective archive.

    The default sorting order of multi-objective records is lexicographic
    based on the objective value vector.

    >>> import numpy as npx
    >>> mr1 = MORecord("xxx", npx.array([1, 2, 3], int))
    >>> print(mr1.x)
    xxx
    >>> print(mr1.fs)
    [1 2 3]
    >>> print(mr1)
    fs=1;2;3, x=xxx
    >>> mr2 = MORecord("yyy", npx.array([1, 2, 1], int))
    >>> print(mr2.x)
    yyy
    >>> print(mr2.fs)
    [1 2 1]
    >>> mr1 < mr2
    False
    >>> mr2 < mr1
    True
    """

    def __init__(self, x: Any, fs: np.ndarray) -> None:
        """
        Create a multi-objective record.

        :param x: the point in the search space
        :param fs: the vector of objective values
        """
        if x is None:
            raise TypeError("x must not be None")
        #: the point in the search space
        self.x: Final[Any] = x
        if not isinstance(fs, np.ndarray):
            raise type_error(fs, "fs", np.ndarray)
        #: the vector of objective values
        self.fs: Final[np.ndarray] = fs

    def __lt__(self, other) -> bool:
        """
        Compare for sorting.

        :param other: the other record

        >>> import numpy as npx
        >>> r1 = MORecord("a", npx.array([1, 1, 1]))
        >>> r2 = MORecord("b", npx.array([1, 1, 1]))
        >>> r1 < r2
        False
        >>> r2 < r1
        False
        >>> r2 = MORecord("b", npx.array([1, 1, 2]))
        >>> r1 < r2
        True
        >>> r2 < r1
        False
        >>> r1 = MORecord("a", npx.array([2, 1, 1]))
        >>> r1 < r2
        False
        >>> r2 < r1
        True
        """
        return lexicographic(self.fs, other.fs) < 0

    def __str__(self):
        """
        Get the string representation of this record.

        :returns: the string representation of this record

        >>> import numpy as npx
        >>> r = MORecord(4, npx.array([1, 2, 3]))
        >>> print(r)
        fs=1;2;3, x=4
        """
        return f"fs={array_to_str(self.fs)}, x={self.x}"


class MOArchivePruner(Component):
    """A strategy for pruning an archive of solutions."""

    def prune(self, archive: list[MORecord], n_keep: int, size: int) -> None:
        """
        Prune an archive.

        After invoking this method, the first `n_keep` entries in `archive`
        are selected to be preserved. The remaining entries
        (at indices `n_keep...len(archive)-1`) can be deleted.

        Pruning therefore is basically just a method of sorting the archive
        according to a preference order of solutions. It will not delete any
        element from the list. The caller can do that afterwards if she wants.

        This base class just provides a simple FIFO scheme.

        :param archive: the archive, i.e., a list of tuples of solutions and
            their objective vectors
        :param n_keep: the number of solutions to keep
        :param size: the current size of the archive
        """
        if size > n_keep:
            n_delete: Final[int] = size - n_keep
            move_to_end: Final[list[MORecord]] = archive[:n_delete]
            archive[0:n_keep] = archive[n_delete:size]
            archive[size - n_delete:size] = move_to_end

    def __str__(self):
        """
        Get the name of this archive pruning strategy.

        :returns: the name of this archive pruning strategy
        """
        return "fifo"


def check_mo_archive_pruner(pruner: Any) -> MOArchivePruner:
    """
    Check whether an object is a valid instance of :class:`MOArchivePruner`.

    :param pruner: the multi-objective archive pruner
    :return: the object
    :raises TypeError: if `pruner` is not an instance of
        :class:`MOArchivePruner`

    >>> check_mo_archive_pruner(MOArchivePruner())
    fifo
    >>> try:
    ...     check_mo_archive_pruner('A')
    ... except TypeError as te:
    ...     print(te)
    pruner should be an instance of moptipy.api.mo_archive.\
MOArchivePruner but is str, namely 'A'.
    >>> try:
    ...     check_mo_archive_pruner(None)
    ... except TypeError as te:
    ...     print(te)
    pruner should be an instance of moptipy.api.mo_archive.\
MOArchivePruner but is None.
    """
    if isinstance(pruner, MOArchivePruner):
        return pruner
    raise type_error(pruner, "pruner", MOArchivePruner)



"""Evaluate the results of the example experiment."""
import argparse
import os.path as pp
from math import log2
from re import Match, Pattern
from re import compile as _compile
from statistics import median
from typing import Callable, Final, Iterable, cast

from moptipy.api.logging import KEY_LAST_IMPROVEMENT_TIME_MILLIS, KEY_TOTAL_FES
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import TIME_UNIT_FES, TIME_UNIT_MILLIS
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import EndStatistics
from moptipy.evaluation.statistics import KEY_MEAN_ARITH
from moptipy.evaluation.tabulate_end_results import (
    DEFAULT_ALGORITHM_INSTANCE_STATISTICS,
    DEFAULT_ALGORITHM_SUMMARY_STATISTICS,
    command_column_namer,
    tabulate_end_results,
)
from moptipy.evaluation.tabulate_result_tests import tabulate_result_tests
from moptipy.examples.jssp.experiment import ALGORITHMS, INSTANCES
from moptipy.examples.jssp.instance import Instance
from moptipy.examples.jssp.plots import (
    plot_end_makespans,
    plot_end_makespans_over_param,
    plot_progresses,
    plot_stat_gantt_charts,
)
from moptipy.spaces.permutations import Permutations
from moptipy.utils.console import logger
from moptipy.utils.help import argparser
from moptipy.utils.lang import EN
from moptipy.utils.logger import SCOPE_SEPARATOR, sanitize_name
from moptipy.utils.path import Path
from moptipy.utils.strings import (
    beautify_float_str,
    name_str_to_num,
    num_to_str,
)
from moptipy.utils.types import type_error

#: The letter mu
LETTER_M: Final[str] = "\u03BC"
#: The letter lambda
LETTER_L: Final[str] = "\u03BB"

#: the pre-defined instance sort keys
__INST_SORT_KEYS: Final[dict[str, int]] = {
    __n: __i for __i, __n in enumerate(INSTANCES)
}

#: the name of the mu+1 EA
NAME_EA_MU_PLUS_1: Final[str] = f"{LETTER_M}+1_ea"
#: the name of the mu+mu EA
NAME_EA_MU_PLUS_MU: Final[str] = f"{LETTER_M}+{LETTER_M}_ea"
#: the name of the mu+ld(mu) EA
NAME_EA_MU_PLUS_LOG_MU: Final[str] = f"{LETTER_M}+log\u2082{LETTER_M}_ea"
#: the name of the mu+sqrt(mu) EA
NAME_EA_MU_PLUS_SQRT_MU: Final[str] = f"{LETTER_M}+\u221A{LETTER_M}_ea"


def __make_all_names() -> tuple[str, ...]:
    """
    Create an immutable list of all algorithm names.

    :returns: the tuple with all algorithm names
    """
    inst = Instance.from_resource("demo")
    space = Permutations.with_repetitions(inst.jobs, inst.machines)
    return tuple([str(a(inst, space)) for a in ALGORITHMS])


#: the list of all algorithm names from the experiment
ALL_NAMES: Final[tuple[str, ...]] = __make_all_names()
del __make_all_names


def ea_family(name: str) -> str:
    """
    Name the evolutionary algorithm setup.

    :param name: the full name
    :returns: the short name of the family
    """
    if name.startswith("ea_"):
        ss: Final[list[str]] = name.split("_")
        ll: Final[int] = len(ss)
        if ll == 4:
            lambda_ = int(ss[2])
            if lambda_ == 1:
                return NAME_EA_MU_PLUS_1
            mu = int(ss[1])
            if lambda_ == int(log2(mu)):
                return NAME_EA_MU_PLUS_LOG_MU
            if lambda_ == round(mu ** 0.5):
                return NAME_EA_MU_PLUS_SQRT_MU
            ratio = lambda_ // mu
            if ratio * mu == lambda_:
                if ratio == 1:
                    return NAME_EA_MU_PLUS_MU
                return f"{LETTER_M}+{ratio}{LETTER_M}_ea"
        elif (ll == 6) and (ss[-2] == "gap") and (ss[-1] == "swap2"):
            return f"{ss[1]}+{ss[2]}_ea_br"
    raise ValueError(f"Invalid EA name {name!r}.")


def sa_family(name: str) -> str:
    """
    Name the simulated annealing setup.

    :param name: the full name
    :returns: the short name of the family
    """
    if name.startswith("sa_") and name.endswith("_swap2"):
        return f"sa_T\u2080_" \
               f"{beautify_float_str(name_str_to_num(name.split('_')[2]))}"
    raise ValueError(f"Invalid SA name {name!r}.")


def ma_family(name: str) -> str:
    """
    Compute the Memetic Algorithm family without the ls steps.

    :param name: the algorithm name
    :returns: the short name of the family
    """
    ss: Final[list[str]] = name.split("_")
    s = "rls" if name.startswith("marls") else "sa"
    return f"{ss[1]}+{ss[2]}_ma{s}"


def __make_algo_names() -> tuple[dict[str, int], dict[str, str]]:
    """
    Create the algorithm sort keys and name decode.

    :returns: the algorithm sort keys and name decode
    """
    names: list[str] = list(ALL_NAMES)
    names_new = list(ALL_NAMES)

    def __eacr(mtch: Match) -> str:
        """Transform the EA name with crossover."""
        br = name_str_to_num(mtch.group(3))
        if not 0.0 < br < 1.0:
            raise ValueError(f"{mtch} has invalid br={br}.")
        a = log2(br)
        b = int(a)
        c = log2(1.0 - br)
        d = int(c)
        q = f"(1-2^{d})" if abs(c - d) < abs(a - b) else f"2^{b}"
        q2 = num_to_str(br)
        if len(q2) <= (len(q) + 1):
            q = q2
        return f"{mtch.group(1)}+{mtch.group(2)}_ea_{q}"

    def __sa(mtch: Match) -> str:
        """Transform the SA name."""
        if mtch.group(0) == "sa_exp16_1em6_swap2":
            return "sa"
        return f"sa_{mtch.group(1)}_{mtch.group(2)}e-{mtch.group(3)}"

    def __ma(mtch: Match, suffix: str) -> str:
        """Transform the MA name."""
        if mtch.group(0) == "ma_2_2_1048576_gap_sa_exp16_5d1em6_swap2":
            return "2_masa_20"
        if mtch.group(0) == "marls_8_8_16_gap_swap2":
            return "8_marls_4"
        ls = name_str_to_num(mtch.group(3))
        if not 0 < ls < 1_000_000_000:
            raise ValueError(f"{mtch} has invalid ls={ls}.")
        e2 = int(log2(ls))
        if ls != (2 ** e2):
            raise ValueError(
                f"{mtch} has invalid ls={ls} since {e2}=log2({ls}).")
        e2s = f"2^{e2}"
        lss = str(ls)
        if len(lss) <= len(e2s):
            e2s = lss
        return f"{mtch.group(1)}+{mtch.group(2)}_ma{suffix}_{e2s}"

    # fix some names using regular expressions
    namer: dict[str, str] = {}
    used_names: set[str] = set(names)
    for pattern, repl in [("hc_swap2", "hc"), ("hcr_32768_swap2", "hcr"),
                          ("hc_swapn", "hcn"), ("hcr_65536_swapn", "hcrn"),
                          ("rls_swap2", "rls"), ("rls_swapn", "rlsn"),
                          ("ea_([0-9]+)_([0-9]+)_swap2", "\\1+\\2_ea"),
                          ("ea_([0-9]+)_([0-9]+)_(0d[0-9]+)_gap_swap2",
                           __eacr),
                          ("sa_exp([0-9]+)_([0-9])em([0-9])_swap2",
                           __sa),
                          ("marls_([0-9]+)_([0-9]+)_([0-9]+)_gap_swap2",
                           lambda mm: __ma(mm, "rls")),
                          ("ma_([0-9]+)_([0-9]+)_([0-9]+)_gap_sa_exp.*",
                           lambda mm: __ma(mm, "sa"))]:
        re: Pattern = _compile(pattern)
        found = False
        for s in names:
            m: Match = re.match(s)
            if m is not None:
                ns = re.sub(cast(str | Callable[[Match], str], repl), s)
                if (ns is None) or (len(ns) <= 0):
                    raise ValueError(f"{s!r} -> {ns!r}?")
                if ns == s:
                    continue
                found = True
                os = namer.get(s, None)
                if os is not None:
                    if os == ns:
                        continue
                    raise ValueError(f"{s!r} -> {ns!r}, {os!r}?")
                if ns in used_names:
                    raise ValueError(f"Already got {ns!r}.")
                namer[s] = ns
                names_new.insert(names_new.index(s), ns)
        if not found:
            raise ValueError(f"did not find {pattern!r}.")

    # fix the basic EA families
    ea1p1: Final[int] = names_new.index("ea_1_1_swap2")
    ea_families: dict[str, int] = {
        NAME_EA_MU_PLUS_MU: ea1p1,
        NAME_EA_MU_PLUS_SQRT_MU: ea1p1,
        NAME_EA_MU_PLUS_LOG_MU: ea1p1,
        NAME_EA_MU_PLUS_1: ea1p1}
    for n in names:
        if n.startswith("ea_"):
            try:
                fam = ea_family(n)
            except ValueError:
                continue
            if fam not in ea_families:
                if fam in used_names:
                    raise ValueError(f"duplicated ea family {fam!r}.")
                used_names.add(fam)
                ea_families[fam] = names_new.index(n)
            else:
                ea_families[fam] = min(ea_families[fam], names_new.index(n))
    for i, n in sorted([(n[1], n[0]) for n in ea_families.items()],
                       reverse=True, key=lambda a: a[0]):
        names_new.insert(i, n)

    # fix the general EA names
    for name in names:
        if name.startswith("generalEa_"):
            n = name[9:]
            has_fitness: bool = False
            for fl, fs in (("_direct", "D"), ("_rank", "R")):
                if fl in n:
                    n = n.replace(fl, fs)
                    has_fitness = True
            if not has_fitness:
                n = f"RI{n}"
            n = n.replace("_tour", "t").replace("_best", "b")\
                .replace("_rndNoRep", "n").replace("_fpsus", "f")\
                .replace("_0d125_gap_swap2", "")
            if "_4_4" in n:
                n = "ea4" + n.replace("_4_4", "")
            elif "_32_32" in n:
                n = "ea32" + n.replace("_32_32", "")
            else:
                raise ValueError(f"invalid general EA: {name!r}.")
            os = namer.get(name, None)
            if os is not None:
                if os == n:
                    continue
                raise ValueError(f"{name!r} -> {n!r}, {os!r}?")
            if n in used_names:
                raise ValueError(f"Already got {n!r}.")
            namer[name] = n
            names_new.insert(names_new.index(name), n)

    # do sa families
    sa_done: set[str] = set()
    for name in names:
        if name.startswith("sa_"):
            n = sa_family(name)
            if n in sa_done:
                continue
            sa_done.add(n)
            names_new.insert(names_new.index(name), n)

    # do ma families
    ma_done: set[str] = set()
    for name in names:
        if name.startswith(("ma_", "marls_")):
            n = ma_family(name)
            if n in ma_done:
                continue
            ma_done.add(n)
            names_new.insert(names_new.index(name), n)

    return {__n: __i for __i, __n in enumerate(names_new)}, namer


#: the pre-defined algorithm sort keys and name map
__ALGO_SORT_KEYS, __ALGO_NAME_MAP = __make_algo_names()


def instance_sort_key(name: str) -> int:
    """
    Get the instance sort key for a given instance name.

    :param name: the instance name
    :returns: the sort key
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __INST_SORT_KEYS:
        return __INST_SORT_KEYS[name]
    return 1000


def algorithm_sort_key(name: str) -> int:
    """
    Get the algorithm sort key for a given algorithm name.

    :param name: the algorithm name
    :returns: the sort key
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __ALGO_SORT_KEYS:
        return __ALGO_SORT_KEYS[name]
    return 1000


def algorithm_namer(name: str) -> str:
    """
    Rename algorithm setups.

    :param name: the algorithm's original name
    :returns: the new name of the algorithm
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __ALGO_NAME_MAP:
        return __ALGO_NAME_MAP[name]
    return name


def compute_end_results(results_dir: str,
                        dest_dir: str) -> Path:
    """
    Get the end results, compute them if necessary.

    :param results_dir: the results directory
    :param dest_dir: the destination directory
    :returns: the path to the end results file.
    """
    dest: Final[Path] = Path.directory(dest_dir)
    results_file: Final[Path] = dest.resolve_inside("end_results.txt")
    if results_file.is_file():
        return results_file
    results: Final[list[EndResult]] = []

    source: Final[Path] = Path.directory(results_dir)
    logger(f"loading end results from path {source!r}.")
    EndResult.from_logs(source, results.append)
    if not results:
        raise ValueError(f"Could not find any logs in {source!r}.")
    results.sort()
    logger(f"found {len(results)} log files in path {source!r}, storing "
           f"them in file {results_file!r}.")
    rf: Path = EndResult.to_csv(results, results_file)
    if rf != results_file:
        raise ValueError(
            f"results file should be {results_file!r}, but is {rf!r}.")
    results_file.enforce_file()
    logger(f"finished writing file {results_file!r}.")
    return results_file


def get_end_results(
        file: str,
        insts: None | set[str] | Callable[[str], bool] = None,
        algos: None | set[str] | Callable[[str], bool] = None) \
        -> list[EndResult]:
    """
    Get a specific set of end results..

    :param file: the end results file
    :param insts: only these instances will be included if this parameter is
        provided
    :param algos: only these algorithms will be included if this parameter is
        provided
    """

    def __filter(er: EndResult,
                 ins=None if insts is None else
                 insts.__contains__ if isinstance(insts, set)
                 else insts,
                 alg=None if algos is None else
                 algos.__contains__ if isinstance(algos, set)
                 else algos) -> bool:
        if (ins is not None) and (not ins(er.instance)):
            return False
        if (alg is not None) and (not alg(er.algorithm)):
            return False
        return True

    col: Final[list[EndResult]] = []
    EndResult.from_csv(file=file, consumer=col.append, filterer=__filter)
    if len(col) <= 0:
        raise ValueError(
            f"no end results for instances {insts} and algorithms {algos}.")
    return col


def compute_end_statistics(end_results_file: str,
                           dest_dir: str) -> Path:
    """
    Get the end result statistics, compute them if necessary.

    :param end_results_file: the end results file
    :param dest_dir: the destination directory
    :returns: the path to the end result statistics file.
    """
    dest: Final[Path] = Path.directory(dest_dir)
    stats_file: Final[Path] = dest.resolve_inside("end_statistics.txt")
    if stats_file.is_file():
        return stats_file

    results: Final[list[EndResult]] = get_end_results(end_results_file)
    if len(results) <= 0:
        raise ValueError("end results cannot be empty")
    stats: Final[list[EndStatistics]] = []
    EndStatistics.from_end_results(results, stats.append)
    if len(stats) <= 0:
        raise ValueError("end result statistics cannot be empty")
    stats.sort()

    sf: Path = EndStatistics.to_csv(stats, stats_file)
    if sf != stats_file:
        raise ValueError(f"stats file should be {stats_file!r} but is {sf!r}")
    stats_file.enforce_file()
    logger(f"finished writing file {stats_file!r}.")
    return stats_file


def table(end_results: Path, algos: list[str], dest: Path,
          swap_stats: Iterable[tuple[str, str]] | None = None) -> None:
    """
    Tabulate the end results.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    :param swap_stats: the statistics to swap out
    """
    EN.set_current()
    n: Final[str] = algorithm_namer(algos[0])
    algo_inst_stat: list | tuple = DEFAULT_ALGORITHM_INSTANCE_STATISTICS
    algo_sum_stat: list | tuple = DEFAULT_ALGORITHM_SUMMARY_STATISTICS
    if swap_stats is not None:
        algo_inst_stat = list(algo_inst_stat)
        algo_sum_stat = list(algo_sum_stat)
        for old, swap in swap_stats:
            found: bool = False
            for lst in [algo_inst_stat, algo_sum_stat]:
                for idx, elem in enumerate(lst):
                    if elem == old:
                        found = True
                        lst[idx] = swap
                        break
            if not found:
                raise ValueError(f"did not find {old!r} in {str(lst)!r}.")

    tabulate_end_results(
        end_results=get_end_results(end_results, algos=set(algos)),
        file_name=sanitize_name(f"end_results_{n}"), dir_name=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        col_namer=command_column_namer,
        algorithm_namer=algorithm_namer,
        algorithm_instance_statistics=algo_inst_stat,
        algorithm_summary_statistics=algo_sum_stat,
        use_lang=False)


def tests(end_results: Path, algos: list[str], dest: Path) -> None:
    """
    Tabulate the end result tests.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    """
    EN.set_current()
    n: Final[str] = algorithm_namer(algos[0])
    tabulate_result_tests(
        end_results=get_end_results(end_results, algos=set(algos)),
        file_name=sanitize_name(f"tests_{n}"), dir_name=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        algorithm_namer=algorithm_namer,
        use_lang=False)


def makespans(end_results: Path, algos: list[str], dest: Path,
              x_label_location: float = 1.0) -> None:
    """
    Plot the end makespans.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    :param x_label_location: the location of the label of the x-axis
    """
    n: Final[str] = algorithm_namer(algos[0])
    plot_end_makespans(
        end_results=get_end_results(end_results, algos=set(algos)),
        name_base=sanitize_name(f"makespan_scaled_{n}"), dest_dir=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        algorithm_namer=algorithm_namer,
        x_label_location=x_label_location)


def gantt(end_results: Path, algo: str, dest: Path, source: Path,
          best: bool = False,
          insts: Iterable[str] | None = None) -> None:
    """
    Plot the median Gantt charts.

    :param end_results: the path to the end results
    :param algo: the algorithm
    :param dest: the directory
    :param source: the source directory
    :param best: should we plot the best instance only (or, otherwise, the
        median)
    :param insts: the instances to use
    """
    n: Final[str] = algorithm_namer(algo)
    plot_stat_gantt_charts(
        get_end_results(end_results, algos={algo},
                        insts=None if insts is None else set(insts)),
        name_base=sanitize_name(f"best_gantt_{n}" if best else f"gantt_{n}"),
        dest_dir=dest,
        results_dir=source,
        instance_sort_key=instance_sort_key,
        statistic=cast(
            Callable[[Iterable[int | float]], int | float],
            min if best else median))


def progress(algos: list[str], dest: Path, source: Path,
             log: bool = True, millis: bool = True) -> None:
    """
    Plot the median Gantt charts.

    :param algos: the algorithms
    :param dest: the directory
    :param source: the source directory
    :param log: is the time logarithmically scaled?
    :param millis: is the time measured in milliseconds?
    """
    n: str = f"progress_{algorithm_namer(algos[0])}_"
    if log:
        n = n + "log_"
    if millis:
        unit = TIME_UNIT_MILLIS
        n = n + "T"
    else:
        unit = TIME_UNIT_FES
        n = n + "FEs"
    plot_progresses(results_dir=source,
                    algorithms=algos,
                    name_base=n,
                    dest_dir=dest,
                    log_time=log,
                    time_unit=unit,
                    algorithm_namer=algorithm_namer,
                    instance_sort_key=instance_sort_key,
                    algorithm_sort_key=algorithm_sort_key)


def makespans_over_param(
        end_results: Path,
        selector: Callable[[str], bool],
        x_getter: Callable[[EndStatistics], int | float],
        name_base: str,
        dest_dir: str,
        x_axis: AxisRanger | Callable[[], AxisRanger]
        = AxisRanger,
        x_label: str | None = None,
        algo_getter: Callable[[str], str] | None = None,
        title: str | None = None,
        legend_pos: str = "right",
        title_x: float = 0.5,
        y_label_location: float = 1.0) -> list[Path]:
    """
    Plot the performance over a parameter.

    :param end_results: the end results path
    :param selector: the selector for algorithms
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param x_getter: the function computing the x-value for each statistics
        object
    :param x_axis: the axis ranger
    :param x_label: the x-axis label
    :param algo_getter: the optional algorithm name getter (use `name_base` if
        unspecified)
    :param title: the optional title (use `name_base` if unspecified)
    :param legend_pos: the legend position, set to "right"
    :param title_x: the title x
    :param y_label_location: the y label location
    :returns: the list of generated files
    """

    def _algo_name_getter(es: EndStatistics,
                          n=name_base, g=algo_getter) -> str:
        return n if g is None else g(es.algorithm)

    return plot_end_makespans_over_param(
        end_results=get_end_results(end_results, algos=selector),
        x_getter=x_getter, name_base=sanitize_name(f"{name_base}_results"),
        dest_dir=dest_dir, title=name_base if title is None else title,
        algorithm_getter=_algo_name_getter,  # type: ignore
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        x_axis=x_axis, x_label=x_label,
        plot_single_instances=algo_getter is None,
        legend_pos=legend_pos, title_x=title_x,
        y_label_location=y_label_location)


def evaluate_experiment(results_dir: str = pp.join(".", "results"),
                        dest_dir: str | None = None) -> None:
    """
    Evaluate the experiment.

    :param results_dir: the results directory
    :param dest_dir: the destination directory
    """
    source: Final[Path] = Path.directory(results_dir)
    dest: Final[Path] = Path.path(dest_dir if dest_dir else
                                  pp.join(source, "..", "evaluation"))
    dest.ensure_dir_exists()
    logger(f"Beginning evaluation from {source!r} to {dest!r}.")

    end_results: Final[Path] = compute_end_results(source, dest)
    if not end_results:
        raise ValueError("End results path is empty??")
    end_stats: Final[Path] = compute_end_statistics(end_results, dest)
    if not end_stats:
        raise ValueError("End stats path is empty??")

    logger("Now evaluating the single random sampling algorithm `1rs`.")
    table(end_results, ["1rs"], dest)
    makespans(end_results, ["1rs"], dest)
    gantt(end_results, "1rs", dest, source)

    logger("Now evaluating the multi-random sampling algorithm `rs`.")
    table(end_results, ["rs", "1rs"], dest)
    makespans(end_results, ["rs", "1rs"], dest)
    gantt(end_results, "rs", dest, source)
    progress(["rs"], dest, source)
    progress(["rs"], dest, source, log=False)

    logger("Now evaluating the hill climbing algorithm `hc`.")
    table(end_results, ["hc_swap2", "rs"], dest)
    makespans(end_results, ["hc_swap2", "rs"], dest)
    gantt(end_results, "hc_swap2", dest, source)
    progress(["hc_swap2", "rs"], dest, source)
    progress(["hc_swap2", "rs"], dest, source, millis=False)

    logger("Now evaluating the hill climbing algorithm with "
           "restarts 'hcr' on 'swap2'.")
    makespans_over_param(
        end_results,
        lambda an: an.startswith("hcr_") and an.endswith("_swap2"),
        lambda es: int(es.algorithm.split("_")[1]),
        "hcr_L_swap2", dest,
        lambda: AxisRanger(log_scale=True, log_base=2.0), "L")
    table(end_results, ["hcr_32768_swap2", "hc_swap2", "rs"], dest)
    makespans(end_results, ["hcr_32768_swap2", "hc_swap2", "rs"], dest)
    gantt(end_results, "hcr_32768_swap2", dest, source)
    progress(["hcr_32768_swap2", "hc_swap2", "rs"], dest, source)

    logger("Now evaluating the hill climbing algorithm with 'swapn'.")
    table(end_results, ["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest)
    makespans(end_results, ["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest)
    progress(["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest, source)
    progress(["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest, source,
             millis=False)

    logger("Now evaluating the hill climbing algorithm with "
           "restarts 'hcr' on 'swapn'.")
    makespans_over_param(
        end_results,
        lambda an: an.startswith("hcr_") and an.endswith("_swapn"),
        lambda es: int(es.algorithm.split("_")[1]),
        "hcr_L_swapn", dest,
        lambda: AxisRanger(log_scale=True, log_base=2.0), "L")
    table(end_results, ["hcr_65536_swapn", "hc_swapn",
                        "hcr_32768_swap2"], dest)
    makespans(end_results, ["hcr_65536_swapn", "hc_swapn",
                            "hcr_32768_swap2"], dest)
    progress(["hcr_65536_swapn", "hc_swapn",
              "hcr_32768_swap2"], dest, source)
    tests(end_results, ["hcr_65536_swapn", "hcr_32768_swap2",
                        "hc_swapn"], dest)

    logger("Now evaluating the RLS algorithm with 'swap2' and 'swapn'.")
    table(end_results, ["rls_swapn", "rls_swap2",
                        "hcr_32768_swap2", "hcr_65536_swapn"], dest)
    tests(end_results, ["rls_swapn", "rls_swap2", "hcr_32768_swap2"], dest)
    makespans(end_results, ["rls_swapn", "rls_swap2", "hcr_32768_swap2",
                            "hcr_65536_swapn"], dest)
    gantt(end_results, "rls_swap2", dest, source)
    progress(["rls_swapn", "rls_swap2", "hcr_32768_swap2",
              "hcr_65536_swapn"], dest, source)
    gantt(end_results, "rls_swap2", dest, source, True, ["ta70"])

    logger("Now evaluating EA without crossover.")

    makespans_over_param(
        end_results=end_results,
        selector=lambda n: n.startswith("ea_") and n.split("_")[3] == "swap2",
        x_getter=lambda es: int(es.algorithm.split("_")[1]),
        name_base="ea_no_cr",
        algo_getter=ea_family,
        title=f"{LETTER_M}+{LETTER_L}_ea", x_label=LETTER_M,
        title_x=0.8,
        x_axis=AxisRanger(log_scale=True, log_base=2.0),
        legend_pos="upper center",
        dest_dir=dest)
    lims: Final[str] = f"{KEY_LAST_IMPROVEMENT_TIME_MILLIS}{SCOPE_SEPARATOR}" \
                       f"{KEY_MEAN_ARITH}"
    totfes: Final[str] = f"{KEY_TOTAL_FES}{SCOPE_SEPARATOR}{KEY_MEAN_ARITH}"
    table(end_results, ["ea_1_2_swap2", "ea_2_2_swap2", "ea_2_4_swap2",
                        "ea_512_512_swap2", "rls_swap2"], dest,
          swap_stats=[(lims, totfes)])
    tests(end_results, ["ea_1_2_swap2", "ea_2_2_swap2", "rls_swap2"], dest)
    makespans(end_results, ["ea_1_2_swap2", "ea_2_2_swap2",
                            "ea_512_512_swap2", "rls_swap2"], dest, 0.6)
    progress(["ea_1_2_swap2", "ea_2_2_swap2", "ea_2_4_swap2",
              "ea_64_1_swap2", "ea_1024_1024_swap2", "ea_8192_65536_swap2",
              "rls_swap2"],
             dest, source)

    logger("Now evaluating EA with crossover.")
    makespans_over_param(
        end_results=end_results,
        selector=lambda n: n.startswith("ea_") and n.endswith("_gap_swap2"),
        x_getter=lambda er: name_str_to_num(er.algorithm.split("_")[3]),
        name_base="ea_cr",
        algo_getter=ea_family,
        title=f"{LETTER_M}+{LETTER_M}_ea_br", x_label="br",
        x_axis=AxisRanger(log_scale=True, log_base=2.0),
        legend_pos="upper right",
        dest_dir=dest)
    table(end_results, ["ea_2_2_0d001953125_gap_swap2", "ea_2_2_swap2",
                        "rls_swap2"], dest)
    tests(end_results, ["ea_2_2_0d001953125_gap_swap2", "ea_2_2_swap2",
                        "rls_swap2"], dest)
    progress(["ea_2_2_0d001953125_gap_swap2", "ea_2_2_swap2", "rls_swap2",
              "ea_256_256_swap2", "ea_256_256_0d25_gap_swap2",
              "ea_32_32_swap2", "ea_32_32_0d125_gap_swap2"],
             dest, source)

    logger("Now evaluating the general EA.")
    for s in ["_4_4", "_32_32"]:
        selected = [y for y in ALL_NAMES if (s in y)
                    and y.startswith("generalEa_")]
        selected.append("rls_swap2")
        selected.append(f"ea{s}_0d125_gap_swap2")
        progress(selected, dest, source, millis=False)

    selected = [y for y in ALL_NAMES if y.startswith("generalEa_")]
    selected.append("rls_swap2")
    selected.append("ea_4_4_0d125_gap_swap2")
    selected.append("ea_32_32_0d125_gap_swap2")
    table(end_results, selected, dest, swap_stats=[(lims, totfes)])

    logger("now preparing to evaluate simulated annealing")
    EN.set_current()
    stats = ["bestF.sd", "lastImprovementFE.med", "totalFEs.med"]
    tabulate_end_results(
        end_results=get_end_results(end_results, algos={"rls_swap2"}),
        file_name="end_stats_rls_2", dir_name=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        col_namer=command_column_namer,
        algorithm_namer=algorithm_namer,
        algorithm_instance_statistics=stats,
        algorithm_summary_statistics=stats,
        put_lower_bound=False,
        use_lang=False)

    logger("now evaluating simulated annealing")
    makespans_over_param(
        end_results=end_results,
        selector=lambda n: n.startswith("sa_") and n.endswith("_swap2") and (
            "1em7" not in n),
        x_getter=lambda er: name_str_to_num(er.algorithm.split("_")[1][3:]),
        name_base="sa_exp",
        algo_getter=sa_family,
        title="sa_T\u2080_\u03b5", x_label="T\u2080",
        x_axis=AxisRanger(log_scale=True, log_base=2.0),
        legend_pos="upper center",
        title_x=0.8,
        dest_dir=dest)
    table(end_results, ["sa_exp16_1em6_swap2", "rls_swap2"], dest)
    tests(end_results, ["sa_exp16_1em6_swap2", "rls_swap2"], dest)
    makespans(end_results, ["sa_exp16_1em6_swap2", "rls_swap2"], dest, 0.72)
    gantt(end_results, "sa_exp16_1em6_swap2", dest, source)
    progress(["sa_exp16_1em6_swap2", "sa_exp16_1em7_swap2",
              "sa_exp16_1em5_swap2", "rls_swap2"], dest, source)
    gantt(end_results, "sa_exp16_1em6_swap2", dest, source, True, ["orb06"])

    logger("now evaluating memetic algorithm")
    table(end_results, ["ma_2_2_1048576_gap_sa_exp16_5d1em6_swap2",
                        "marls_8_8_16_gap_swap2",
                        "sa_exp16_1em6_swap2"], dest,
          swap_stats=[(lims, totfes)])
    makespans_over_param(
        end_results=end_results,
        selector=lambda n: n.startswith("ma"),
        x_getter=lambda er: name_str_to_num(er.algorithm.split("_")[3]),
        name_base="ma_",
        algo_getter=ma_family,
        title=f"{LETTER_M}+{LETTER_L}_ma{{rls|sa}}", x_label="ls_fes",
        x_axis=AxisRanger(log_scale=True, log_base=2.0),
        legend_pos="lower center",
        title_x=0.8,
        y_label_location=0.1,
        dest_dir=dest)
    progress(["marls_8_8_16_gap_swap2",
              "ma_2_2_1048576_gap_sa_exp16_5d1em6_swap2",
              "sa_exp16_1em6_swap2", "ea_2_2_swap2",
              "ea_8_8_swap2"], dest, source)
    makespans(end_results, ["ma_2_2_1048576_gap_sa_exp16_5d1em6_swap2",
                            "sa_exp16_1em6_swap2"], dest, 0.72)

    logger(f"Finished evaluation from {source!r} to {dest!r}.")


# Evaluate experiment if run as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = argparser(
        __file__, "Evaluate the results of the JSSP example experiment",
        "This experiment evaluates all the results of the JSSP example"
        " experiment and creates the figures and tables of the "
        "'Optimization Algorithms' book (see "
        "http://thomasweise.github.io/oa).")
    parser.add_argument(
        "source", nargs="?", default="./results", type=Path.path,
        help="the directory with the results of the JSSP experiment")
    parser.add_argument(
        "dest", type=Path.path, nargs="?", default="./evaluation/",
        help="the directory to write the evaluation results to")
    args: Final[argparse.Namespace] = parser.parse_args()
    evaluate_experiment(results_dir=args.source, dest_dir=args.dest)



"""Plot the end results over a parameter."""
from math import isfinite
from typing import Any, Callable, Final, Iterable, cast

from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import F_NAME_SCALED
from moptipy.evaluation.end_statistics import EndStatistics
from moptipy.evaluation.statistics import KEY_MEAN_GEOM
from moptipy.evaluation.styler import Styler
from moptipy.utils.lang import Lang
from moptipy.utils.logger import SCOPE_SEPARATOR
from moptipy.utils.types import type_error


def __make_y_label(y_dim: str) -> str:
    """
    Make the y label.

    :param y_dim: the y dimension
    :returns: the y label
    """
    dotidx: Final[int] = y_dim.find(SCOPE_SEPARATOR)
    if dotidx > 0:
        y_dimension: Final[str] = y_dim[:dotidx]
        y_stat: Final[str] = y_dim[dotidx + 1:]
        return Lang.translate_func(y_stat)(y_dimension)
    return Lang.translate(y_dim)


def __make_y_axis(y_dim: str) -> AxisRanger:
    """
    Make the y axis.

    :param y_dim: the y dimension
    :returns: the y axis
    """
    dotidx: Final[int] = y_dim.find(SCOPE_SEPARATOR)
    if dotidx > 0:
        y_dim = y_dim[:dotidx]
    return AxisRanger.for_axis(y_dim)


def plot_end_statistics_over_param(
        data: Iterable[EndStatistics],
        figure: Axes | Figure,
        x_getter: Callable[[EndStatistics], int | float],
        y_dim: str = f"{F_NAME_SCALED}{SCOPE_SEPARATOR}{KEY_MEAN_GEOM}",
        algorithm_getter: Callable[[EndStatistics], str | None] =
        lambda es: es.algorithm,
        instance_getter: Callable[[EndStatistics], str | None] =
        lambda es: es.instance,
        x_axis: AxisRanger | Callable[[], AxisRanger] = AxisRanger,
        y_axis: AxisRanger | Callable[[str], AxisRanger] =
        __make_y_axis,
        legend: bool = True,
        legend_pos: str = "upper right",
        distinct_colors_func: Callable[[int], Any] = pd.distinct_colors,
        distinct_line_dashes_func: Callable[[int], Any] =
        pd.distinct_line_dashes,
        importance_to_line_width_func: Callable[[int], float] =
        pd.importance_to_line_width,
        importance_to_font_size_func: Callable[[int], float] =
        pd.importance_to_font_size,
        x_grid: bool = True,
        y_grid: bool = True,
        x_label: str | None = None,
        x_label_inside: bool = True,
        x_label_location: float = 0.5,
        y_label: None | str | Callable[[str], str] = __make_y_label,
        y_label_inside: bool = True,
        y_label_location: float = 1.0,
        instance_priority: float = 0.666,
        algorithm_priority: float = 0.333,
        stat_priority: float = 0.0,
        instance_sort_key: Callable[[str], Any] = lambda x: x,
        algorithm_sort_key: Callable[[str], Any] = lambda x: x,
        instance_namer: Callable[[str], str] = lambda x: x,
        algorithm_namer: Callable[[str], str] = lambda x: x,
        stat_sort_key: Callable[[str], str] = lambda x: x,
        color_algorithms_as_fallback_group: bool = True) -> Axes:
    """
    Plot a series of end result statistics over a parameter.

    :param data: the iterable of EndStatistics
    :param figure: the figure to plot in
    :param x_getter: the function computing the x-value for each statistics
        object
    :param y_dim: the dimension to be plotted along the y-axis
    :param algorithm_getter: the algorithm getter
    :param instance_getter: the instance getter
    :param x_axis: the x_axis ranger
    :param y_axis: the y_axis ranger
    :param legend: should we plot the legend?
    :param legend_pos: the legend position
    :param distinct_colors_func: the function returning the palette
    :param distinct_line_dashes_func: the function returning the line styles
    :param importance_to_line_width_func: the function converting importance
        values to line widths
    :param importance_to_font_size_func: the function converting importance
        values to font sizes
    :param x_grid: should we have a grid along the x-axis?
    :param y_grid: should we have a grid along the y-axis?
    :param x_label: the label for the x-axi or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param x_label_location: the location of the x-axis label
    :param y_label: a callable returning the label for the y-axis, a label
        string, or `None` if no label should be put
    :param y_label_inside: put the y-axis label inside the plot (so that
        it does not consume additional horizontal space)
    :param y_label_location: the location of the y-axis label
    :param instance_priority: the style priority for instances
    :param algorithm_priority: the style priority for algorithms
    :param stat_priority: the style priority for statistics
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param instance_namer: the name function for instances receives an
        instance ID and returns an instance name; default=identity function
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :param stat_sort_key: the sort key function for statistics
    :param color_algorithms_as_fallback_group: if only a single group of data
        was found, use algorithms as group and put them in the legend
    :returns: the axes object to allow you to add further plot elements
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(data, Iterable):
        raise type_error(data, "data", Iterable)
    if not isinstance(figure, Axes | Figure):
        raise type_error(figure, "figure", (Axes, Figure))
    if not callable(x_getter):
        raise type_error(x_getter, "x_getter", call=True)
    if not isinstance(y_dim, str):
        raise type_error(y_dim, "y_dim", str)
    if len(y_dim) <= 0:
        raise ValueError(f"invalid y-dimension {y_dim!r}")
    if not callable(instance_getter):
        raise type_error(instance_getter, "instance_getter", call=True)
    if not callable(algorithm_getter):
        raise type_error(algorithm_getter, "algorithm_getter", call=True)
    if not isinstance(legend, bool):
        raise type_error(legend, "legend", bool)
    if not isinstance(legend_pos, str):
        raise type_error(legend_pos, "legend_pos", str)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(distinct_line_dashes_func):
        raise type_error(
            distinct_line_dashes_func, "distinct_line_dashes_func", call=True)
    if not callable(importance_to_font_size_func):
        raise type_error(importance_to_font_size_func,
                         "importance_to_font_size_func", call=True)
    if not isinstance(x_grid, bool):
        raise type_error(x_grid, "x_grid", bool)
    if not isinstance(y_grid, bool):
        raise type_error(y_grid, "y_grid", bool)
    if not ((x_label is None) or isinstance(x_label, str)):
        raise type_error(x_label, "x_label", (str, None))
    if not isinstance(x_label_inside, bool):
        raise type_error(x_label_inside, "x_label_inside", bool)
    if not isinstance(x_label_location, float):
        raise type_error(x_label_location, "x_label_location", float)
    if not ((y_label is None) or callable(y_label)
            or isinstance(y_label, str)):
        raise type_error(y_label, "y_label", (str, None), call=True)
    if not isinstance(y_label_inside, bool):
        raise type_error(y_label_inside, "y_label_inside", bool)
    if not isinstance(y_label_location, float):
        raise type_error(y_label_location, "y_label_location", float)
    if not isinstance(instance_priority, float):
        raise type_error(instance_priority, "instance_priority", float)
    if not isfinite(instance_priority):
        raise ValueError(f"instance_priority cannot be {instance_priority}.")
    if not isinstance(algorithm_priority, float):
        raise type_error(algorithm_priority, "algorithm_priority", float)
    if not isfinite(algorithm_priority):
        raise ValueError(f"algorithm_priority cannot be {algorithm_priority}.")
    if not isinstance(stat_priority, float):
        raise type_error(stat_priority, "stat_priority", float)
    if not isfinite(stat_priority):
        raise ValueError(f"stat_priority cannot be {stat_priority}.")
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)
    if not callable(stat_sort_key):
        raise type_error(stat_sort_key, "stat_sort_key", call=True)
    if not callable(instance_namer):
        raise type_error(instance_namer, "instance_namer", call=True)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)
    if not isinstance(color_algorithms_as_fallback_group, bool):
        raise type_error(color_algorithms_as_fallback_group,
                         "color_algorithms_as_fallback_group", bool)

    # the getter for the dimension value
    y_getter: Final[Callable[[EndStatistics], int | float]] \
        = cast(Callable[[EndStatistics], int | float],
               EndStatistics.getter(y_dim))
    if not callable(y_getter):
        raise type_error(y_getter, "y-getter", call=True)

    # set up the axis rangers
    if callable(x_axis):
        x_axis = x_axis()
    if not isinstance(x_axis, AxisRanger):
        raise type_error(x_axis, "x_axis", AxisRanger)

    if callable(y_axis):
        y_axis = y_axis(y_dim)
    if not isinstance(y_axis, AxisRanger):
        raise type_error(y_axis, "y_axis", AxisRanger)

    # First, we try to find groups of data to plot together in the same
    # color/style. We distinguish progress objects from statistical runs.
    instances: Final[Styler] = Styler(
        none_name=Lang.translate("all_insts"),
        priority=instance_priority,
        namer=instance_namer,
        name_sort_function=instance_sort_key)
    algorithms: Final[Styler] = Styler(
        none_name=Lang.translate("all_algos"),
        namer=algorithm_namer,
        priority=algorithm_priority, name_sort_function=algorithm_sort_key)

    # we now extract the data: x -> algo -> inst -> y
    dataset: Final[dict[str | None, dict[
        str | None, dict[int | float, int | float]]]] = {}
    for endstat in data:
        if not isinstance(endstat, EndStatistics):
            raise type_error(endstat, "element in data", EndStatistics)
        x_value = x_getter(endstat)
        if not isinstance(x_value, int | float):
            raise type_error(x_value, "x-value", (int, float))
        _algo = algorithm_getter(endstat)
        if not ((_algo is None) or isinstance(_algo, str)):
            raise type_error(_algo, "algorithm name", None, call=True)
        _inst = instance_getter(endstat)
        if not ((_inst is None) or isinstance(_inst, str)):
            raise type_error(_algo, "instance name", None, call=True)
        y_value = y_getter(endstat)
        if not isinstance(y_value, int | float):
            raise type_error(y_value, "y-value", (int, float))
        if _algo in dataset:
            _dataset = dataset[_algo]
        else:
            dataset[_algo] = _dataset = {}
        if _inst in _dataset:
            __dataset = _dataset[_inst]
        else:
            _dataset[_inst] = __dataset = {}
        if x_value in __dataset:
            raise ValueError(
                f"combination x={x_value}, algo={_algo!r}, inst={_inst!r} "
                f"already known as value {__dataset[x_value]}, cannot assign "
                f"value {y_value}.")
        __dataset[x_value] = y_value
        x_axis.register_value(x_value)
        y_axis.register_value(y_value)
        algorithms.add(_algo)
        instances.add(_inst)
    del data, y_getter, x_getter, x_value, y_value

    if len(dataset) <= 0:
        raise ValueError("no data found?")

    def __set_importance(st: Styler) -> None:
        none = 1
        not_none = 0
        none_lw = importance_to_line_width_func(none)
        not_none_lw = importance_to_line_width_func(not_none)
        st.set_line_width(lambda p: [none_lw if i <= 0 else not_none_lw
                                     for i in range(p)])

    # determine the style groups
    groups: list[Styler] = []
    instances.finalize()
    algorithms.finalize()

    if instances.count > 1:
        groups.append(instances)
    if algorithms.count > 1:
        groups.append(algorithms)

    if len(groups) > 0:
        groups.sort()
        groups[0].set_line_color(distinct_colors_func)

        if len(groups) > 1:
            groups[1].set_line_dash(distinct_line_dashes_func)
    elif color_algorithms_as_fallback_group:
        algorithms.set_line_color(distinct_colors_func)
        groups.append(algorithms)

    # If we only have <= 2 groups, we can mark None and not-None values with
    # different importance.
    if instances.has_none and (instances.count > 1):
        __set_importance(instances)
    elif algorithms.has_none and (algorithms.count > 1):
        __set_importance(algorithms)

    # we will collect all lines to plot in plot_list
    plot_list: list[dict] = []
    for algo in algorithms.keys:
        _dataset = dataset[algo]
        for inst in instances.keys:
            if inst not in _dataset:
                raise ValueError(f"instance {inst!r} not in dataset"
                                 f" for algorithm {algo!r}.")
            __dataset = _dataset[inst]
            style = pd.create_line_style()
            style["x"] = x_vals = sorted(__dataset.keys())
            style["y"] = [__dataset[x] for x in x_vals]
            for g in groups:
                g.add_line_style(inst if g is instances else algo, style)
            plot_list.append(style)
    del dataset, _dataset, __dataset

    # now we have all data, let's move to the actual plotting
    font_size_0: Final[float] = importance_to_font_size_func(0)

    # set up the graphics area
    axes: Final[Axes] = pu.get_axes(figure)
    axes.tick_params(axis="x", labelsize=font_size_0)
    axes.tick_params(axis="y", labelsize=font_size_0)

    # draw the grid
    if x_grid or y_grid:
        grid_lwd = importance_to_line_width_func(-1)
        if x_grid:
            axes.grid(axis="x", color=pd.GRID_COLOR, linewidth=grid_lwd)
        if y_grid:
            axes.grid(axis="y", color=pd.GRID_COLOR, linewidth=grid_lwd)

    # plot the lines
    for line in plot_list:
        axes.step(where="post", **line)
    del plot_list

    # make sure that we can see the maximum of the parameters
    x_axis.pad_detected_range(pad_max=True)

    x_axis.apply(axes, "x")
    y_axis.apply(axes, "y")

    if legend:
        handles: list[Artist] = []

        for g in groups:
            g.add_to_legend(handles.append)
            g.has_style = False

        if instances.has_style:
            instances.add_to_legend(handles.append)
        if algorithms.has_style:
            algorithms.add_to_legend(handles.append)

        if len(handles) > 0:
            axes.legend(loc=legend_pos,
                        handles=handles,
                        labelcolor=[art.color if hasattr(art, "color")
                                    else pd.COLOR_BLACK for art in handles],
                        fontsize=font_size_0)

    pu.label_axes(axes=axes,
                  x_label=x_label,
                  x_label_inside=x_label_inside,
                  x_label_location=x_label_location,
                  y_label=y_label(y_dim) if callable(y_label) else y_label,
                  y_label_inside=y_label_inside,
                  y_label_location=y_label_location,
                  font_size=font_size_0)
    return axes



"""Print a help screen."""

import argparse
import os.path
import sys
from typing import Final

from moptipy.utils.path import _canonicalize_path
from moptipy.utils.types import type_error
from moptipy.version import __version__

#: The default argument parser for moptipy executables.
__DEFAULT_ARGUMENTS: Final[argparse.ArgumentParser] = argparse.ArgumentParser(
    epilog="Copyright\u00a0\u00a9\u00a02022\u00a0Thomas\u00a0WEISE, "
           "GNU\u00a0GENERAL\u00a0PUBLIC\u00a0LICENSE\u00a0Version\u00a03,"
           "\u00a029\u00a0June\u00a02007, "
           "https://thomasweise.github.io/moptipy, "
           "tweise@hfuu.edu.cn,\u00a0tweise@ustc.edu.cn",
    add_help=False,
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
)
__DEFAULT_ARGUMENTS.add_argument(
    "--version", action="version", version=__version__)


def __get_python_interpreter_short() -> str:
    """
    Get the python interpreter.

    :returns: the fully-qualified path
    """
    inter: Final[str] = _canonicalize_path(sys.executable)
    bn = os.path.basename(inter)
    if bn.startswith("python3."):
        bn2 = bn[:7]
        interp2 = os.path.join(os.path.dirname(inter), bn2)
        if os.path.exists(interp2) and os.path.isfile(interp2) \
                and (_canonicalize_path(interp2) == inter):
            return bn2
    return bn


#: The python interpreter in short form.
__INTERPRETER_SHORT: Final[str] = __get_python_interpreter_short()
del __get_python_interpreter_short

#: the base path of the moptipy package
__BASE_PATH: Final[str] = _canonicalize_path(os.path.dirname(
    os.path.dirname(os.path.dirname(_canonicalize_path(__file__))))) + os.sep


def __get_prog(file: str) -> str:
    """
    Get the program as to be displayed by the help screen.

    The result of this function applied to the `__file__` special
    variable should be put into the `prog` argument of the constructor
    of :class:`argparse.ArgumentParser`.

    :param file: the calling python script
    :return: the program string
    """
    if not isinstance(file, str):
        raise type_error(file, "file", str)

    # get the module minus the base path and extension
    module: str = _canonicalize_path(file)
    end: int = len(module)
    start: int = 0
    if module.endswith(".py"):
        end -= 3
    if module.startswith(__BASE_PATH):
        start += len(__BASE_PATH)
    module = module[start:end].replace(os.sep, ".")

    return f"{__INTERPRETER_SHORT} -m {module}"


def argparser(file: str, description: str,
              epilog: str) -> argparse.ArgumentParser:
    """
    Create an argument parser with default settings.

    :param file: the `__file__` special variable of the calling script
    :param description: the description string
    :param epilog: the epilogue string
    :returns: the argument parser

    >>> ap = argparser(__file__, "This is a test program.", "This is a test.")
    >>> isinstance(ap, argparse.ArgumentParser)
    True
    >>> "Copyright" in ap.epilog
    True
    """
    if not isinstance(file, str):
        raise type_error(file, "file", str)
    if len(file) <= 3:
        raise ValueError(f"invalid file={file!r}.")
    if not isinstance(description, str):
        raise type_error(description, "description", str)
    if len(description) <= 12:
        raise ValueError(f"invalid description={description!r}.")
    if not isinstance(epilog, str):
        raise type_error(epilog, "epilog", str)
    if len(epilog) <= 10:
        raise ValueError(f"invalid epilog={epilog!r}.")
    return argparse.ArgumentParser(
        parents=[__DEFAULT_ARGUMENTS], prog=__get_prog(file),
        description=description.strip(),
        epilog=f"{epilog.strip()} {__DEFAULT_ARGUMENTS.epilog}",
        formatter_class=__DEFAULT_ARGUMENTS.formatter_class)



"""
Parsers for structured log data produced by the `moptipy` experiment API.

The `moptipy` :class:`~moptipy.api.execution.Execution` and experiment-running
facility (:func:`~moptipy.api.experiment.run_experiment`) uses the class
:class:`~moptipy.utils.logger.Logger` from module :mod:`~moptipy.utils.logger`
to produce log files complying with
https://thomasweise.github.io/moptipy/#log-files.

Here we provide a skeleton for parsing such log files in form of the class
:class:`~LogParser`. It works similar to SAX-XML parsing in that the data
is read is from files and methods that consume the data are invoked. By
overwriting these methods, we can do useful things with the data.

For example in module :mod:`~moptipy.evaluation.end_results`, the method
:meth:`~moptipy.evaluation.end_results.EndResult.from_logs` can load
:class:`~moptipy.evaluation.end_results.EndResult` records from the logs
and the method :meth:`~moptipy.evaluation.progress.Progress.from_logs` in
module :mod:`~moptipy.evaluation.progress` reads the whole
:class:`~moptipy.evaluation.progress.Progress` that the algorithms make
over time.
"""

from os import listdir
from os.path import basename, dirname, isdir, isfile, join
from typing import Final

from moptipy.api import logging
from moptipy.utils.console import logger
from moptipy.utils.logger import COMMENT_CHAR, SECTION_END, SECTION_START
from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.path import Path
from moptipy.utils.strings import PART_SEPARATOR, sanitize_name


class LogParser:
    """
    A log parser can parse a log file and separate the sections.

    The log parser is designed to load data from text files generated
    by :class:`~moptipy.utils.logger.FileLogger`. It can also recursively
    parse directories.
    """

    def __init__(self,
                 print_begin_end: bool = True,
                 print_file_start: bool = False,
                 print_file_end: bool = False,
                 print_dir_start: bool = True,
                 print_dir_end: bool = True):
        """
        Initialize the log parser.

        :param print_begin_end: log to stdout when starting and ending
            a recursive directory parsing process
        :param print_file_start: log to stdout when opening a file
        :param print_file_end: log to stdout when closing a file
        :param print_dir_start: log to stdout when entering a directory
        :param print_dir_end: log to stdout when leaving a directory
        """
        self.__print_begin_end: Final[bool] = print_begin_end
        self.__print_file_start: Final[bool] = print_file_start
        self.__print_file_end: Final[bool] = print_file_end
        self.__print_dir_start: Final[bool] = print_dir_start
        self.__print_dir_end: Final[bool] = print_dir_end
        self.__depth: int = 0

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def start_dir(self, path: Path) -> bool:
        """
        Enter a directory to parse all files inside.

        This method is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`.
        If it returns `True`, every sub-directory inside of it will be passed
        to :meth:`~moptipy.evaluation.log_parser.LogParser.start_dir`
        and every file will be passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file`.
        Only if `True` is returned,
        :meth:`~moptipy.evaluation.log_parser.LogParser.end_dir` will be
        invoked and its return value will be the return value of
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`. If `False`
        is returned, then
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir` will return
        immediately and return `True`.

        :param path: the path of the directory
        :return: `True` if all the files and sub-directories inside the
            directory should be processed, `False` if this directory should
            be skipped and parsing should continue with the next sibling
            directory
        """
        del path
        return True

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def end_dir(self, path: Path) -> bool:
        """
        Enter a directory to parse all files inside.

        This method is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`.
        If it returns `True`, every sub-directory inside of it will be passed
        to :meth:`~moptipy.evaluation.log_parser.LogParser.start_dir`
        and every file will be passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file`.

        :param path: the path of the directory
        :return: `True` if all the files and sub-directories inside the
            directory should be processed, `False` if this directory should
            be skipped and parsing should continue with the next sibling
            directory
        """
        del path
        return True

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def start_file(self, path: Path) -> bool:
        """
        Decide whether to start parsing a file.

        This method is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file`. If it
        returns `True`, then we will open and parse the file. If it returns
        `False`, then the fill will not be parsed and
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` will
        return `True` immediately.

        :param path: the file path
        :return: `True` if the file should be parsed, `False` if it should be
            skipped (and
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` should
            return `True`).
        """
        return path.endswith(logging.FILE_SUFFIX)

    # noinspection PyMethodMayBeStatic
    def start_section(self, title: str) -> bool:
        """
        Start a section.

        If this method returns `True`, then all the lines of text of the
        section `title` will be read and together passed to
        :meth:`~moptipy.evaluation.log_parser.LogParser.lines`.
        If this method returns `False`, then the section will be skipped
        and we fast-forward to the next section, if any, or to the call
        of :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`.

        :param title: the section title
        :return: `True` if the section data should be loaded and passed to
            :meth:`lines`, `False` of the section can be skipped. In that
            case, we will fast-forward to the next
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_section`.
        """
        if not title:
            raise ValueError(f"Title cannot be empty, but is {title!r}.")
        if title.startswith(logging.ERROR_SECTION_PREFIX):
            raise ValueError(f"Encountered error section {title!r}.")
        return False

    # noinspection PyUnusedLocal
    # noinspection PyMethodMayBeStatic
    def lines(self, lines: list[str]) -> bool:
        """
        Consume all the lines from a section.

        This method receives the complete text of a section, where all lines
        are separated and put into one list `lines`. Each line is stripped
        from whitespace and comments, empty lines are omitted.
        If this method returns `True`, we will continue parsing the file and
        move to the next section, if any, or directly to the end of the file
        parsing process (and call
        :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`) if there is
        no more section in the file.

        :param lines: the lines to consume
        :return: `True` if further parsing is necessary and the next section
            should be fed to
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_section`,
            `False` if the parsing
            process can be terminated, in which case we will fast-forward to
            :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`
        """
        del lines
        return True

    # noinspection PyMethodMayBeStatic
    def end_file(self) -> bool:
        """
        End a file.

        This method is invoked when we have reached the end of the current
        file. Its return value, `True` or `False`, will then be returned
        by :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file`, which
        is the entry point for the file parsing process.

        :return: the return value to be returned by
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file`
        """
        return True

    def parse_file(self, path: str) -> bool:
        """
        Parse the contents of a file.

        This method first calls the function
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file` to see
        whether the file should be parsed. If
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file` returns
        `True`, then the file is parsed. If
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_file` returns
        `False`, then this method returns `False` directly.
        If the file is parsed, then
        :meth:`~moptipy.evaluation.log_parser.LogParser.start_section` will
        be invoked for each section (until the parsing is finished) and
        :meth:`~moptipy.evaluation.log_parser.LogParser.lines` for
        each section content (if requested). At the end,
        :meth:`~moptipy.evaluation.log_parser.LogParser.end_file` is invoked.

        This method can either be called directly or is called by
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir`. In the
        latter case, if
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` returned
        `True`, the next file in the current directory will be parsed. If it
        returns `False`, then no file located in the current directory will be
        parsed, while other directories and/or sub-directories will still be
        processed.

        :param path: the file to parse
        :return: the return value received from invoking
            :meth:`~moptipy.evaluation.log_parser.LogParser.end_file`
        """
        file: Final[Path] = Path.file(path)

        retval: bool
        try:
            retval = self.start_file(file)
        except BaseException as be:
            raise ValueError(f"Error when starting file {file!r}") from be

        if retval:
            if self.__print_file_start:
                logger(f"parsing file {file!r}.")
        else:
            if self.__print_file_start:
                logger(f"skipping file {file!r}.")
            return True

        lines: list[str] = []
        buffer: list[str] = []
        state: int = 0
        wants_section: bool = False
        sec_end: str = ""
        section: str = ""
        sect_start: Final[str] = SECTION_START
        sect_end: Final[str] = SECTION_END
        cmt_chr: Final[str] = COMMENT_CHAR

        index: int = 0
        with file.open_for_read() as handle:
            while True:

                # get the next line
                if index >= len(buffer):
                    try:
                        buffer = handle.readlines(128)
                    except BaseException as be:
                        raise ValueError(
                            f"Error when reading lines from file {file!r} "
                            f"while in section {section!r}."
                            if state == 1 else
                            f"Error when reading lines from file {file!r}.") \
                            from be
                    if (buffer is None) or (len(buffer) <= 0):
                        break
                    index = 0

                orig_cur = buffer[index]
                index += 1

                # strip next line from comments and white space
                cur = orig_cur.strip()
                if len(cur) <= 0:
                    continue

                i = cur.find(cmt_chr)
                if i >= 0:
                    cur = cur[:i].strip()
                    if len(cur) <= 0:
                        continue

                if state in (0, 2):
                    if not cur.startswith(sect_start):
                        raise ValueError("Line should start with "
                                         f"{sect_start!r} but is "
                                         f"{orig_cur!r} in file {file!r}.")
                    section = cur[len(sect_start):]
                    if len(section) <= 0:
                        raise ValueError(
                            "Section title cannot be empty in "
                            f"{file!r}, but encountered {orig_cur!r}.")
                    state = 1
                    sec_end = sect_end + section
                    wants_section = self.start_section(section)
                elif state == 1:
                    if cur == sec_end:
                        state = 2
                        if wants_section:
                            try:
                                do_next = self.lines(lines)
                            except BaseException as be:
                                raise ValueError(
                                    "Error when processing section "
                                    f"{section!r} in file {file!r}.") from be
                            lines.clear()
                            if not do_next:
                                break
                    elif wants_section:
                        lines.append(cur)

        if state == 0:
            raise ValueError(f"Log file {file!r} contains no section.")
        if state == 1:
            raise ValueError(f"Log file {file!r} ended before"
                             f"encountering {sec_end!r}.")

        try:
            retval = self.end_file()
        except BaseException as be:
            raise ValueError("Error when ending section parsing "
                             f"of file {file!r}.") from be
        if self.__print_file_end:
            logger(f"finished parsing file {file!r}.")
        return retval

    def parse_dir(self, path: str) -> bool:
        """
        Recursively parse the given directory.

        :param path: the path to the directory
        :return: `True` either if
            :meth:`~moptipy.evaluation.log_parser.LogParser.start_dir`
            returned `False` or
            :meth:`~moptipy.evaluation.log_parser.LogParser.end_dir` returned
            `True`, `False` otherwise
        """
        folder: Final[Path] = Path.directory(path)

        if self.__depth <= 0:
            if self.__depth == 0:
                if self.__print_begin_end:
                    logger("beginning recursive parsing of "
                           f"directory {folder!r}.")
            else:
                raise ValueError(
                    f"Depth must be >= 0, but is {self.__depth}??")
        self.__depth += 1

        if not self.start_dir(folder):
            if self.__print_dir_start:
                logger(f"skipping directory {folder!r}.")
            return True
        if self.__print_dir_start:
            logger(f"entering directory {folder!r}.")

        do_files = True
        do_dirs = True
        for subpath in listdir(folder):
            sub = Path.path(join(folder, subpath))
            if isfile(sub):
                if do_files and (not self.parse_file(sub)):
                    logger(f"will parse no more files in {folder!r}.")
                    if not do_dirs:
                        break
                    do_files = False
            elif isdir(sub) and do_dirs and (not self.parse_dir(sub)):
                logger("will parse no more sub-directories "
                       f"of {folder!r}.")
                if not do_files:
                    break
                do_dirs = False

        retval = self.end_dir(folder)
        if self.__print_dir_end:
            logger(f"finished parsing directory {folder!r}.")

        self.__depth -= 1
        if self.__depth <= 0:
            if self.__depth == 0:
                if self.__print_begin_end:
                    logger(f"finished recursive parsing of "
                           f"directory {folder!r}.")
            else:
                raise ValueError(
                    f"Depth must be >= 0, but is {self.__depth}??")

        return retval

    def parse(self, path: str) -> bool:
        """
        Parse either a directory or a file.

        If `path` identifies a file,
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` is invoked
        and its result is returned. If `path` identifies a directory, then
        :meth:`~moptipy.evaluation.log_parser.LogParser.parse_dir` is invoked
        and its result is returned.

        :param path: a path identifying either a directory or a file.
        :return: the result of the appropriate parsing routing
        :raises ValueError: if `path` does not identify a directory or file
        """
        npath = Path.path(path)
        if isfile(npath):
            return self.parse_file(npath)
        if isdir(npath):
            return self.parse_dir(npath)
        raise ValueError(
            f"Path {npath} is neither a file nor a directory?")


class ExperimentParser(LogParser):
    """A log parser following our pre-defined experiment structure."""

    def __init__(self):
        """Initialize the experiment parser."""
        super().__init__(print_begin_end=True, print_dir_start=True)

        #: The name of the algorithm to which the current log file belongs.
        self.algorithm: str | None = None

        #: The name of the instance to which the current log file belongs.
        self.instance: str | None = None

        #: The random seed of the current log file.
        self.rand_seed: int | None = None

    def start_file(self, path: Path) -> bool:
        """
        Decide whether to start parsing a file and setup meta-data.

        :param path: the file path
        :return: `True` if the file should be parsed, `False` if it should be
            skipped (and
            :meth:`~moptipy.evaluation.log_parser.LogParser.parse_file` should
            return `True`).
        """
        if not super().start_file(path):
            return False

        inst_dir = dirname(path)
        algo_dir = dirname(inst_dir)
        self.instance = sanitize_name(basename(inst_dir))
        self.algorithm = sanitize_name(basename(algo_dir))

        start = f"{self.algorithm}{PART_SEPARATOR}" \
                f"{self.instance}{PART_SEPARATOR}0x"
        base = basename(path)
        if not base.startswith(start):
            raise ValueError(
                f"File name of {path!r} should start with {start!r}.")
        self.rand_seed = rand_seed_check(int(
            base[len(start):(-len(logging.FILE_SUFFIX))], base=16))

        return True

    def end_file(self) -> bool:
        """Finalize parsing a file."""
        self.rand_seed = None
        self.algorithm = None
        self.instance = None
        return super().end_file()



"""
A simple implementation of a (mu+lambda) Memetic Algorithm.

A memetic algorithm (:class:`~moptipy.algorithms.so.ma.MA`) works similar to a
(mu+lambda) :class:`~moptipy.algorithms.so.ea.EA`, but it refines all results
of the search operators with a local search
:attr:`~moptipy.algorithms.so.ma.MA.ls` executed for
:attr:`~moptipy.algorithms.so.ma.MA.ls_fes` objective function evaluations.
It also only employs the nullary search operator (to create the initial random
solutions) and the binary search operator (to combine two selected parental
solutions).

This is the general form of the Memetic Algorithm of the type
"EA+something else." A specialized version combining an
:mod:`~moptipy.algorithms.so.ea` with
:mod:`~moptipy.algorithms.so.rls` can be found in
:mod:`~moptipy.algorithms.so.marls`.

This Memetic Algorithm implementation begins by sampling
:attr:`~moptipy.algorithms.so.ma.MA.mu`
solutions using the nullary search operation
:attr:`~moptipy.api.algorithm.Algorithm0.op0`. Each of these initial solutions
is used as a starting point of a local search
:attr:`~moptipy.algorithms.so.ma.MA.ls`, which is executed for
:attr:`~moptipy.algorithms.so.ma.MA.ls_fes` objective function evaluations.
In each iteration, it then uses the
:attr:`~moptipy.algorithms.so.ma.MA.mu` existing solutions as input for
the binary search operator :attr:`~moptipy.algorithms.so.ma.MA.op2` to create
:attr:`~moptipy.algorithms.so.ma.MA.lambda_` new solutions, each of which is
again used as a starting point of a local search
:attr:`~moptipy.algorithms.so.ma.MA.ls` executed for
:attr:`~moptipy.algorithms.so.ma.MA.ls_fes` objective function evaluations.
The results of the local searches enter the population and in the next
iteration, the :attr:`~moptipy.algorithms.so.ma.MA.mu` best solutions of the
:attr:`~moptipy.algorithms.so.ma.MA.mu` +
:attr:`~moptipy.algorithms.so.ma.MA.lambda_` ones in the population are
retained.

Due to the :class:`~moptipy.api.process.Process` and
:mod:`~moptipy.api.subprocesses` API of `moptipy`, you can use almost arbitrary
algorithms as local search :attr:`~moptipy.algorithms.so.ma.MA.ls`. The only
requirement is that is a subclass of
:class:`~moptipy.api.algorithm.Algorithm0` and uses
it uses an instance of
:class:`moptipy.operators.op0_forward.Op0Forward` as nullary search operator
(:attr:`~moptipy.api.algorithm.Algorithm0.op0`).
This allows the MA to set a solution as starting point for the inner algorithm
:attr:`~moptipy.algorithms.so.ma.MA.ls`.

It should be noted that it is by no means required that the inner algorithm
:attr:`~moptipy.algorithms.so.ma.MA.ls` needs to be a local search. Any
algorithm that fulfills the above requirements could be used. For example, if
we were conducting numerical optimization, it would be totally OK to plug an
instance of the Sequential Least Squares Programming
(:class:`~moptipy.algorithms.so.vector.scipy.SLSQP`) algorithm into the
memetic algorithm&hellip;

Further reading on how to realize using one algorithm as a sub-algorithm
of another one can be found in the documentation of
:func:`~moptipy.api.subprocesses.from_starting_point`,
:func:`~moptipy.api.subprocesses.for_fes`, and
:class:`moptipy.operators.op0_forward.Op0Forward`.

1. Pablo Moscato. *On Evolution, Search, Optimization, Genetic Algorithms and
   Martial Arts: Towards Memetic Algorithms.* Caltech Concurrent Computation
   Program Report C3P 826. 1989. Pasadena, CA, USA: California Institute of
   Technology (Caltech), Caltech Concurrent Computation Program (C3P).
   https://www.researchgate.net/publication/2354457
2. Carlos Cotta, Luke Mathieson, and Pablo Moscato. Memetic Algorithms. In
   Rafael Martí, Panos M. Pardalos, and Mauricio G. C. Resende, editors,
   *Handbook of Heuristics.* Part~III: Metaheuristics, pages 607-638. 2018.
   Cham, Switzerland: Springer. ISBN: 978-3-319-07123-7.
   doi: https://doi.org/10.1007/978-3-319-07153-4_29-1
   https://www.researchgate.net/publication/315660932
3. William Eugene Hart, James E. Smith, and Natalio Krasnogor, editors.
   *Recent Advances in Memetic Algorithms.* Studies in Fuzziness and Soft
   Computing (STUDFUZZ), volume 166. 2005. Berlin/Heidelberg, Germany:
   Springer. ISBN: 978-3-540-22904-9.
   doi: https://doi.org/10.1007/3-540-32363-5
4. Ferrante Neri, Carlos Cotta, and Pablo Moscato. *Handbook of Memetic
   Algorithms.* Volume 379 of Studies in Computational Intelligence (SCI).
   2012. Berlin/Heidelberg, Germany: Springer. ISBN: 978-3-642-23246-6
   doi https://doi.org/10.1007/978-3-642-23247-3.
5. L. Darrell Whitley, V. Scott Gordon, and Keith E. Mathias. Lamarckian
   Evolution, The Baldwin Effect and Function Optimization. In Yuval Davidor,
   Hans-Paul Schwefel, and Reinhard Männer, editors, *Proceedings of the Third
   Conference on Parallel Problem Solving from Nature; International
   Conference on Evolutionary Computation (PPSN III),* October 9-14, 1994,
   Jerusalem, Israel, pages 5-15. Volume 866/1994 of Lecture Notes in Computer
   Science, Berlin, Germany: Springer-Verlag GmbH. ISBN 0387584846.
   doi: https://doi.org/10.1007/3-540-58484-6_245.
   https://www.researchgate.net/publication/2521727
"""
from typing import Callable, Final, cast

from numpy.random import Generator

from moptipy.algorithms.so.record import Record
from moptipy.api.algorithm import Algorithm0
from moptipy.api.logging import SCOPE_OP2
from moptipy.api.operators import Op0, Op2, check_op2
from moptipy.api.process import Process
from moptipy.api.subprocesses import for_fes, from_starting_point
from moptipy.operators.op0_forward import Op0Forward
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import PART_SEPARATOR
from moptipy.utils.types import check_int_range, type_error


# start book
class MA(Algorithm0):
    """An MA is a population-based algorithm using binary operators."""

    def solve(self, process: Process) -> None:
        """
        Apply the MA to an optimization problem.

        :param process: the black-box process object
        """
        # initialization of some variables omitted in book for brevity
        # end book
        mu: Final[int] = self.mu  # mu: number of best solutions kept
        mu_plus_lambda: Final[int] = mu + self.lambda_  # size
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        ls_fes: Final[int] = self.ls_fes  # the number of FEs per ls run
        ls_solve: Final[Callable[[Process], None]] = self.ls.solve  # +book
        forward_ls_op0_to: Final[Callable] = cast(  # forward starting
            Op0Forward, self.ls.op0).forward_to  # point of ls to...
        should_terminate: Final[Callable] = process.should_terminate
        r0i: Final[Callable[[int], int]] = cast(  # random integers
            Callable[[int], int], random.integers)
        # start book
        # create list of mu random+ls records and lambda empty records
        lst: Final[list] = [None] * mu_plus_lambda  # pre-allocate list
        f: int | float = 0  # variable to hold objective values
        for i in range(mu_plus_lambda):  # fill list of size mu+lambda
            x = create()  # by creating point in search space
            if i < mu:  # only the first mu records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we stop now?
                    cast(Op0Forward, self.ls.op0).stop_forwarding()  # -book
                    return   # computational budget exhausted -> quit
                with for_fes(process, ls_fes) as s1, \
                        from_starting_point(s1, x, evaluate(x)) as s2:
                    forward_ls_op0_to(s2.get_copy_of_best_x)
                    ls_solve(s2)  # apply local search modifying x
                    f = s2.get_best_f()  # get quality of x
            lst[i] = Record(x, f)  # create and store record

        it: int = 0  # set iteration counter=0 (but immediately increment)
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            it += 1  # step iteration counter
            for oi in range(mu, mu_plus_lambda):  # for all offspring
                if should_terminate():  # should we stop now?
                    cast(Op0Forward, self.ls.op0).stop_forwarding()  # -book
                    return   # computational budget exhausted -> quit
                dest: Record = lst[oi]  # pick destination record
                x = dest.x  # the destination "x" value
                dest.it = it  # remember iteration of solution creation

                sx = lst[r0i(mu)].x  # pick random first source "x"
                sx2 = sx    # second source "x" initially=first sx
                while sx2 is sx:     # until different from sx...
                    sx2 = lst[r0i(mu)].x  # ..get random second "x"
                op2(random, x, sx, sx2)  # apply binary operator
                with for_fes(process, ls_fes) as s1, \
                        from_starting_point(s1, x, evaluate(x)) as s2:
                    forward_ls_op0_to(s2.get_copy_of_best_x)
                    ls_solve(s2)  # apply local search modifying x
                    dest.f = s2.get_best_f()  # get quality of x
            lst.sort()  # best records come first, ties broken by age
            # end book

    def __init__(self, op0: Op0,
                 op2: Op2, ls: Algorithm0,
                 mu: int = 2, lambda_: int = 1,
                 ls_fes: int = 1000,
                 name: str = "ma") -> None:
        """
        Create the Memetic Algorithm (MA).

        :param op0: the nullary search operator
        :param op2: the binary search operator
        :param ls: the local search to apply to each new solution
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        :param ls_fes: the number of FEs (steps) per local search run
        :param name: the base name of the algorithm
        """
        super().__init__(f"{name}{PART_SEPARATOR}{mu}{PART_SEPARATOR}"
                         f"{lambda_}{PART_SEPARATOR}{ls_fes}{PART_SEPARATOR}"
                         f"{op2}{PART_SEPARATOR}{ls}", op0)
        if not isinstance(ls, Algorithm0):
            raise type_error(ls, "ls", Algorithm0)
        if not isinstance(ls.op0, Op0Forward):
            raise type_error(ls.op0, "ls.op0", Op0Forward)
        #: the number of records to survive in each generation
        self.mu: Final[int] = check_int_range(mu, "mu", 1, 1_000_000)
        #: the number of offsprings per generation
        self.lambda_: Final[int] = check_int_range(
            lambda_, "lambda", 1, 1_000_000)
        #: the number of FEs per local search run
        self.ls_fes: Final[int] = check_int_range(
            ls_fes, "ls_fes", 1, 100_000_000)
        #: The binary search operator.
        self.op2: Final[Op2] = check_op2(op2)
        #: the local search algorithm
        self.ls: Final[Algorithm0] = ls

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("mu", self.mu)
        logger.key_value("lambda", self.lambda_)
        logger.key_value("lsFEs", self.ls_fes)
        with logger.scope(SCOPE_OP2) as o:
            self.op2.log_parameters_to(o)
        with logger.scope("ls") as ls:
            self.ls.log_parameters_to(ls)

    def initialize(self) -> None:
        """Initialize the memetic algorithm."""
        super().initialize()
        self.ls.initialize()
        self.op2.initialize()



"""Functions that can be used to test objective functions."""
from math import inf, isfinite
from typing import Any, Callable, Final

from numpy.random import Generator, default_rng

from moptipy.api.objective import Objective, check_objective
from moptipy.api.space import Space
from moptipy.tests.component import validate_component
from moptipy.utils.types import type_error


def validate_objective(
        objective: Objective,
        solution_space: Space | None = None,
        make_solution_space_element_valid:
        Callable[[Generator, Any], Any] | None = lambda _, x: x,
        is_deterministic: bool = True,
        lower_bound_threshold: int | float = -inf,
        upper_bound_threshold: int | float = inf,
        must_be_equal_to: Callable[[Any], int | float] | None = None) -> None:
    """
    Check whether an object is a moptipy objective function.

    :param objective: the objective function to test
    :param solution_space: the solution space
    :param make_solution_space_element_valid: a function that makes an element
        from the solution space valid
    :param bool is_deterministic: is the objective function deterministic?
    :param lower_bound_threshold: the threshold for the lower bound
    :param upper_bound_threshold: the threshold for the upper bound
    :param must_be_equal_to: an optional function that should return the
        exactly same values as the objective function
    :raises ValueError: if `objective` is not a valid
        :class:`~moptipy.api.objective.Objective`
    :raises TypeError: if values of the wrong types are encountered
    """
    if not isinstance(objective, Objective):
        raise type_error(objective, "objective", Objective)
    check_objective(objective)
    validate_component(objective)

    if not (hasattr(objective, "lower_bound")
            and callable(getattr(objective, "lower_bound"))):
        raise ValueError("objective must have method lower_bound.")
    lower: Final[int | float] = objective.lower_bound()
    if not (isinstance(lower, int | float)):
        raise type_error(lower, "lower_bound()", (int, float))
    if (not isfinite(lower)) and (not (lower <= (-inf))):
        raise ValueError(
            f"lower bound must be finite or -inf, but is {lower}.")
    if lower < lower_bound_threshold:
        raise ValueError("lower bound must not be less than "
                         f"{lower_bound_threshold}, but is {lower}.")

    if not (hasattr(objective, "upper_bound")
            and callable(getattr(objective, "upper_bound"))):
        raise ValueError("objective must have method upper_bound.")
    upper: Final[int | float] = objective.upper_bound()
    if not (isinstance(upper, int | float)):
        raise type_error(upper, "upper_bound()", (int, float))
    if (not isfinite(upper)) and (not (upper >= inf)):
        raise ValueError(
            f"upper bound must be finite or +inf, but is {upper}.")
    if upper > upper_bound_threshold:
        raise ValueError(
            f"upper bound must not be more than {upper_bound_threshold}, "
            f"but is {upper}.")

    if lower >= upper:
        raise ValueError("Result of lower_bound() must be smaller than "
                         f"upper_bound(), but got {lower} vs. {upper}.")

    if not (hasattr(objective, "is_always_integer")
            and callable(getattr(objective, "is_always_integer"))):
        raise ValueError("objective must have method is_always_integer.")
    is_int: Final[bool] = objective.is_always_integer()
    if not isinstance(is_int, bool):
        raise type_error(is_int, "is_always_integer()", bool)
    if is_int:
        if isfinite(lower) and (not isinstance(lower, int)):
            raise TypeError(
                f"if is_always_integer()==True, then lower_bound() must "
                f"return int, but it returned {lower}.")
        if isfinite(upper) and (not isinstance(upper, int)):
            raise TypeError(
                f"if is_always_integer()==True, then upper_bound() must "
                f"return int, but it returned {upper}.")

    count: int = 0
    if make_solution_space_element_valid is not None:
        count += 1
    if solution_space is not None:
        count += 1
    if count <= 0:
        return
    if count < 2:
        raise ValueError("either provide both of solution_space and "
                         "make_solution_space_element_valid or none.")

    x = solution_space.create()
    if x is None:
        raise ValueError("solution_space.create() produced None.")
    x = make_solution_space_element_valid(default_rng(), x)
    if x is None:
        raise ValueError("make_solution_space_element_valid() produced None.")
    solution_space.validate(x)

    if not (hasattr(objective, "evaluate")
            and callable(getattr(objective, "evaluate"))):
        raise ValueError("objective must have method evaluate.")
    res = objective.evaluate(x)
    if not (isinstance(res, int | float)):
        raise type_error(res, f"evaluate(x) of {x}", (int, float))

    if (res < lower) or (res > upper):
        raise ValueError(f"evaluate(x) of {x} must return a value in"
                         "[lower_bound(), upper_bound()], but returned "
                         f"{res} vs. [{lower},{upper}].")
    if is_int and (not isinstance(res, int)):
        raise TypeError(
            f"if is_always_integer()==True, then evaluate(x) must "
            f"return int, but it returned {res}.")

    if must_be_equal_to is not None:
        exp = must_be_equal_to(x)
        if exp != res:
            raise ValueError(f"expected to get {exp}, but got {res}.")

    res2 = objective.evaluate(x)
    if not (isinstance(res2, int | float)):
        raise type_error(res2, f"evaluate(x) of {x}", (int, float))

    if (res2 < lower) or (res2 > upper):
        raise ValueError(f"evaluate(x) of {x} must return a value in"
                         "[lower_bound(), upper_bound()], but returned "
                         f"{res2} vs. [{lower},{upper}].")
    if is_int and (not isinstance(res2, int)):
        raise TypeError(
            f"if is_always_integer()==True, then evaluate(x) must "
            f"return int, but it returned {res2}.")

    if is_deterministic and (res != res2):
        raise ValueError(f"evaluating {x} twice yielded the two different "
                         f"results {res} and {res2}!")



"""
Approximate the :class:`~moptipy.evaluation.ecdf.Ecdf` to reach certain goals.

The empirical cumulative distribution function (ECDF) for short illustrates
the fraction of runs that have reached a certain goal over time. Let's say
that you have performed 10 runs of a certain algorithm on a certain problem.
As goal quality, you could define the globally optimal solution quality.
For any point in time, the ECDF then shows how many of these runs have solved
the problem to this goal, to optimality.
Let's say the first run solves the problem after 100 FEs.
Then the ECDF is 0 until 99 FEs and at 100 FEs, it becomes 1/10.
The second fastest run solves the problem after 200 FEs.
The ECDF thus stays 0.1 until 199 FEs and at 200 FEs, it jumps to 0.2.
And so on.
This means that the value of the ECDF is always between 0 and 1.

1. Nikolaus Hansen, Anne Auger, Steffen Finck, Raymond Ros. *Real-Parameter
   Black-Box Optimization Benchmarking 2010: Experimental Setup.*
   Research Report RR-7215, INRIA. 2010. inria-00462481.
   https://hal.inria.fr/inria-00462481/document/
2. Dave Andrew Douglas Tompkins and Holger H. Hoos. UBCSAT: An Implementation
   and Experimentation Environment for SLS Algorithms for SAT and MAX-SAT. In
   *Revised Selected Papers from the Seventh International Conference on
   Theory and Applications of Satisfiability Testing (SAT'04),* May 10-13,
   2004, Vancouver, BC, Canada, pages 306-320. Lecture Notes in Computer
   Science (LNCS), volume 3542. Berlin, Germany: Springer-Verlag GmbH.
   ISBN: 3-540-27829-X. doi: https://doi.org/10.1007/11527695_24.
3. Holger H. Hoos and Thomas Stützle. Evaluating Las Vegas Algorithms -
   Pitfalls and Remedies. In Gregory F. Cooper and Serafín Moral, editors,
   *Proceedings of the 14th Conference on Uncertainty in Artificial
   Intelligence (UAI'98)*, July 24-26, 1998, Madison, WI, USA, pages 238-245.
   San Francisco, CA, USA: Morgan Kaufmann Publishers Inc.
   ISBN: 1-55860-555-X.
"""

from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable

import numpy as np

import moptipy.api.logging as lg
import moptipy.utils.nputils as npu
from moptipy.evaluation._utils import _get_reach_index
from moptipy.evaluation.base import (
    F_NAME_NORMALIZED,
    F_NAME_SCALED,
    KEY_N,
    MultiRun2DData,
)
from moptipy.evaluation.progress import Progress
from moptipy.utils.console import logger
from moptipy.utils.lang import Lang
from moptipy.utils.logger import (
    COMMENT_CHAR,
    CSV_SEPARATOR,
    KEY_VALUE_SEPARATOR,
)
from moptipy.utils.path import Path
from moptipy.utils.strings import num_to_str
from moptipy.utils.types import check_int_range, type_error

#: The number of instances.
KEY_N_INSTS: Final[str] = f"{KEY_N}Insts"
#: The objective dimension name.
KEY_F_NAME: Final[str] = "fName"


@dataclass(frozen=True, init=False, order=True)
class Ecdf(MultiRun2DData):
    """The ECDF data."""

    #: The ECDF data function
    ecdf: np.ndarray
    #: The number of instances over which the ERT-ECDF is computed.
    n_insts: int
    #: The goal value, or None if different goals were used for different
    #: instances
    goal_f: int | float | None

    def __init__(self,
                 algorithm: str | None,
                 n: int,
                 n_insts: int,
                 time_unit: str,
                 f_name: str,
                 goal_f: int | float | None,
                 ecdf: np.ndarray):
        """
        Create the ECDF function.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param n: the total number of runs
        :param n_insts: the total number of instances
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        :param goal_f: the goal value, or `None` if different goals were used
            for different instances
        :param numpy.ndarray ecdf: the ert-ecdf matrix
        """
        super().__init__(algorithm, None, n, time_unit, f_name)
        object.__setattr__(
            self, "n_insts", check_int_range(n_insts, "n_insts", 1, self.n))

        if goal_f is not None:
            if not isinstance(goal_f, int | float):
                raise type_error(goal_f, "goal_f", (int, float))
            if not isfinite(goal_f):
                raise ValueError(f"Invalid goal_f {goal_f}.")
        object.__setattr__(self, "goal_f", goal_f)

        if not isinstance(ecdf, np.ndarray):
            raise type_error(ecdf, "ecdf", np.ndarray)
        ecdf.flags.writeable = False

        time: Final[np.ndarray] = ecdf[:, 0]
        ll = time.size
        if ll < 2:
            raise ValueError("Must have at least two points in "
                             f"ecdf curve , but encountered {ll}.")
        if not npu.is_all_finite(time[:-1]):
            raise ValueError("Non-last Ert-based time-values must be finite, "
                             f"but encountered {time}.")
        if np.any(time[1:] <= time[:-1]):
            raise ValueError("Time data must be strictly increasing,"
                             f"but encountered {time}.")

        prb: Final[np.ndarray] = ecdf[:, 1]
        if not npu.is_all_finite(prb):
            raise ValueError(
                f"All ECDF values must be finite, but encountered {prb}.")
        if (ll > 2) and np.any(prb[1:-1] <= prb[:-2]):
            raise ValueError("ECDF data must be strictly increasing,"
                             f"but encountered {prb}.")
        if prb[0] < 0:
            raise ValueError(f"First ECDF element cannot be {prb[0]}.")
        if prb[ll - 1] > 1:
            raise ValueError(f"Last ECDF element cannot be {prb[ll - 1]}.")
        object.__setattr__(self, "ecdf", ecdf)

    def time_label(self) -> str:
        """
        Get the time label for x-axes.

        :return: the time key
        """
        return Lang.translate(self.time_unit)

    def _time_key(self) -> str:
        """
        Get the time key.

        :return: the time key
        """
        return self.time_unit

    def to_csv(self, file: str,
               put_header: bool = True) -> Path:
        """
        Store a :class:`Ecdf` record in a CSV file.

        :param file: the file to generate
        :param put_header: should we put a header with meta-data?
        :return: the fully resolved file name
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing ECDF to CSV file {path!r}.")

        with path.open_for_write() as out:
            sep: Final[str] = CSV_SEPARATOR
            if put_header:
                kv: Final[str] = KEY_VALUE_SEPARATOR
                cmt: Final[str] = COMMENT_CHAR
                if self.algorithm is not None:
                    out.write(
                        f"{cmt} {lg.KEY_ALGORITHM}{kv}{self.algorithm}\n")
                out.write(
                    f"{cmt} {KEY_N}{kv}{self.n}\n")
                out.write(
                    f"{cmt} {KEY_N_INSTS}{kv}{self.n_insts}\n")
                out.write(
                    f"{cmt} {KEY_F_NAME}{kv}{self.f_name}\n")
                if self.goal_f is not None:
                    out.write(
                        f"{cmt} {lg.KEY_GOAL_F}{kv}{self.goal_f}\n")

            out.write(f"{self._time_key()}{sep}ecdf\n")
            for v in self.ecdf:
                out.write(
                    f"{num_to_str(v[0])}{sep}{num_to_str(v[1])}\n")

        logger(f"Done writing ECDF to CSV file {path!r}.")

        path.enforce_file()
        return path

    @staticmethod
    def _compute_times(source: list[Progress],
                       goal: int | float) -> list[float]:
        """
        Compute the times for the given goals.

        :param source: the source array
        :param goal: the goal value
        :return: a list of times
        """
        ret = []
        for pr in source:
            idx = _get_reach_index(pr.f, goal)
            if idx <= 0:
                continue
            ret.append(pr.time[pr.time.size - idx])
        return ret

    # noinspection PyUnusedLocal
    @staticmethod
    def _get_div(n: int, n_insts: int) -> int:
        """
        Get the divisor.

        :param n: the number of runs
        :param n_insts: the number of instances
        :return: the divisor
        """
        del n_insts
        return n

    @classmethod
    def create(cls: type["Ecdf"],
               source: Iterable[Progress],
               goal_f: int | float | Callable | None = None,
               use_default_goal_f: bool = True) -> "Ecdf":
        """
        Create one single Ecdf record from an iterable of Progress records.

        :param source: the set of progress instances
        :param goal_f: the goal objective value
        :param use_default_goal_f: should we use the default lower bounds as
            goals?
        :return: the Ecdf record
        :rtype: Ecdf
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)

        algorithm: str | None = None
        time_unit: str | None = None
        f_name: str | None = None
        inst_runs: dict[str, list[Progress]] = {}
        n: int = 0

        for progress in source:
            if not isinstance(progress, Progress):
                raise type_error(progress, "progress", Progress)
            if n <= 0:
                algorithm = progress.algorithm
                time_unit = progress.time_unit
                f_name = progress.f_name
            else:
                if algorithm != progress.algorithm:
                    algorithm = None
                if time_unit != progress.time_unit:
                    raise ValueError(
                        f"Cannot mix time units {time_unit} "
                        f"and {progress.time_unit}.")
                if f_name != progress.f_name:
                    raise ValueError(f"Cannot mix f-names {f_name} "
                                     f"and {progress.f_name}.")
            n += 1
            if progress.instance in inst_runs:
                inst_runs[progress.instance].append(progress)
            else:
                inst_runs[progress.instance] = [progress]
        del source

        if n <= 0:
            raise ValueError("Did not encounter any progress information.")
        n_insts: Final[int] = len(inst_runs)
        if (n_insts <= 0) or (n_insts > n):
            raise ValueError("Huh?.")

        times: list[float] = []
        goal: int | float | None
        same_goal_f: int | float | None = None
        first: bool = True
        for instance, pl in inst_runs.items():
            goal = None
            if isinstance(goal_f, int | float):
                goal = goal_f
            elif callable(goal_f):
                goal = goal_f(instance)
            if (goal is None) and use_default_goal_f:
                if f_name == F_NAME_SCALED:
                    goal = 1
                elif f_name == F_NAME_NORMALIZED:
                    goal = 0
                else:
                    goal = pl[0].f_standard
                    for pp in pl:
                        if goal != pp.f_standard:
                            raise ValueError("Inconsistent goals: "
                                             f"{goal} and {pp.f_standard}")
            if not isinstance(goal, int | float):
                raise type_error(goal, "goal", (int, float))
            if first:
                same_goal_f = goal
            elif goal != same_goal_f:
                same_goal_f = None
            res = cls._compute_times(pl, goal)
            for t in res:
                if isfinite(t):
                    if t < 0:
                        raise ValueError(f"Invalid ert {t}.")
                    times.append(t)
                elif not (t >= inf):
                    raise ValueError(f"Invalid ert {t}.")
        del inst_runs

        if len(times) <= 0:
            return cls(algorithm, n, n_insts, time_unit,
                       f_name, same_goal_f,
                       np.array([[0, 0], [inf, 0]]))

        times.sort()
        time: list[float] = [0]
        ecdf: list[float] = [0]
        success: int = 0
        div: Final[int] = cls._get_div(n, n_insts)
        ll: int = 0
        for t in times:
            success += 1
            if t > time[ll]:
                time.append(t)
                ecdf.append(success / div)
                ll += 1
            else:
                ecdf[ll] = success / div

        time.append(inf)
        ecdf.append(ecdf[ll])

        return cls(algorithm, n, n_insts,
                   time_unit, f_name,
                   same_goal_f,
                   np.column_stack((np.array(time),
                                    np.array(ecdf))))

    @classmethod
    def from_progresses(cls: type["Ecdf"],
                        source: Iterable[Progress],
                        consumer: Callable[["Ecdf"], Any],
                        f_goal: int | float | Callable | Iterable[
                            int | float | Callable] = None,
                        join_all_algorithms: bool = False) -> None:
        """
        Compute one or multiple ECDFs from a stream of end results.

        :param source: the set of progress instances
        :param f_goal: one or multiple goal values
        :param consumer: the destination to which the new records will be
            passed, can be the `append` method of a :class:`list`
        :param join_all_algorithms: should the Ert-Ecdf be aggregated over all
            algorithms
        """
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(join_all_algorithms, bool):
            raise type_error(join_all_algorithms,
                             "join_all_algorithms", bool)
        if not isinstance(f_goal, Iterable):
            f_goal = [f_goal]

        sorter: dict[str, list[Progress]] = {}
        for er in source:
            if not isinstance(er, Progress):
                raise type_error(er, "progress", Progress)
            key = er.algorithm
            if key in sorter:
                lst = sorter[key]
            else:
                lst = []
                sorter[key] = lst
            lst.append(er)

        if len(sorter) <= 0:
            raise ValueError("source must not be empty")

        keyz = list(sorter.keys())
        keyz.sort()

        for goal in f_goal:
            use_default_goal = goal is None
            for key in keyz:
                consumer(cls.create(sorter[key], goal, use_default_goal))


def get_goal(ecdf: Ecdf) -> int | float | None:
    """
    Get the goal value from the given ecdf instance.

    :param Ecdf ecdf: the ecdf instance
    :return: the goal value
    """
    return ecdf.goal_f


def goal_to_str(goal_f: int | float | None) -> str:
    """
    Transform a goal to a string.

    :param goal_f: the goal value
    :return: the string representation
    """
    return "undefined" if goal_f is None else \
        f"goal: \u2264{num_to_str(goal_f)}"



"""Test all the example code in the project's examples directory."""
import os.path
from typing import Final

from numpy.random import default_rng

from moptipy.utils.console import logger
from moptipy.utils.path import Path
from moptipy.utils.temp import TempDir


def test_examples_in_examples_directory() -> None:
    """Test all the examples in the examples directory."""
    # First, we resolve the directories
    base_dir = Path.directory(os.path.join(os.path.dirname(__file__), "../"))
    examples_dir = Path.directory(base_dir.resolve_inside("examples"))
    logger(
        f"executing all examples from examples directory {examples_dir!r}.")

    wd: Final[str] = os.getcwd()  # get current working directory
    with TempDir.create() as td:  # create temporary working directory
        logger(f"using temp directory {td!r}.")
        os.chdir(td)  # set it as working directory
        files: list[str] = os.listdir(examples_dir)
        logger(f"got {len(files)} potential files")
        default_rng().shuffle(files)  # shuffle the order for randomness
        for name in files:  # find all files in examples
            if name.endswith(".py"):  # if it is a python file
                file: Path = Path.file(examples_dir.resolve_inside(name))
                logger(f"now compiling file {file!r}.")
                code = compile(  # noqa # nosec
                    file.read_all_str(), file, mode="exec")  # noqa # nosec
                logger(f"now executing file {file!r}.")
                exec(code, {})  # execute file  # noqa # nosec
                logger(f"successfully executed example {file!r}.")
    os.chdir(wd)  # go back to current original directory
    logger("finished executing all examples from README.md.")
